from __future__ import absolute_import, division, print_function

import os
from collections import defaultdict, Counter

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt

from glue.external.echo import SelectionCallbackProperty
from glue.external.echo.qt import connect_combo_selection
from glue.core import ComponentID
from glue.core.parse import ParsedComponentLink, ParsedCommand
from glue.utils.qt import load_ui
from glue.core.message import NumericalDataChangedMessage

from glue.dialogs.component_arithmetic.qt.equation_editor import EquationEditorDialog

__all__ = ['ArithmeticEditorWidget']


class ArithmeticEditorWidget(QtWidgets.QDialog):

    data = SelectionCallbackProperty()

    def __init__(self, data_collection=None, initial_data=None, parent=None):

        super(ArithmeticEditorWidget, self).__init__(parent=parent)

        self.ui = load_ui('component_arithmetic.ui', self,
                          directory=os.path.dirname(__file__))

        self.list = self.ui.list_derived_components

        self.data_collection = data_collection

        self._components_derived = defaultdict(list)
        self._components_other = defaultdict(list)
        self._state = defaultdict(dict)

        for data in data_collection:

            # First find all derived components (only ones based on arithmetic
            # expressions)

            self._components_derived[data] = []

            for cid in data.derived_components:
                comp = data.get_component(cid)
                if isinstance(comp.link, ParsedComponentLink):
                    comp_state = {}
                    comp_state['cid'] = cid
                    comp_state['label'] = cid.label
                    comp_state['equation'] = comp.link._parsed
                    self._state[data][cid] = comp_state
                    self._components_derived[data].append(cid)

            # Keep track of all other components

            self._components_other[data] = []

            for cid in data.components:
                if cid not in self._components_derived[data]:
                    self._components_other[data].append(cid)

        # Populate data combo
        ArithmeticEditorWidget.data.set_choices(self, list(self.data_collection))
        ArithmeticEditorWidget.data.set_display_func(self, lambda x: x.label)
        self._connection = connect_combo_selection(self, 'data', self.ui.combosel_data)

        if initial_data is None:
            self.ui.combosel_data.setCurrentIndex(0)
        else:
            self.data = initial_data

        self.ui.combosel_data.currentIndexChanged.connect(self._update_component_lists)
        self._update_component_lists()

        self.ui.button_add_derived.clicked.connect(self._add_derived_component)
        self.ui.button_edit_derived.clicked.connect(self._edit_derived_component)
        self.ui.button_remove_derived.clicked.connect(self._remove_derived_component)

        self.ui.list_derived_components.itemSelectionChanged.connect(self._update_selection_derived)

        self._update_selection_derived()

        self.ui.list_derived_components.itemChanged.connect(self._update_state)
        self.ui.list_derived_components.order_changed.connect(self._update_state)
        self.ui.list_derived_components.itemDoubleClicked.connect(self._edit_derived_component)

        self.ui.button_ok.clicked.connect(self.accept)
        self.ui.button_cancel.clicked.connect(self.reject)

    def _update_selection_derived(self):
        enabled = self.list.selected_cid is not None
        self.button_edit_derived.setEnabled(enabled)
        self.button_remove_derived.setEnabled(enabled)

    def _update_component_lists(self, *args):

        # This gets called when the data is changed and we need to update the
        # components shown in the lists.

        self.list.blockSignals(True)

        mapping = {}
        for cid in self.data.components:
            mapping[cid] = cid.label

        self.list.clear()
        for cid in self._components_derived[self.data]:
            label = self._state[self.data][cid]['label']
            if self._state[self.data][cid]['equation'] is None:
                expression = ''
            else:
                expression = self._state[self.data][cid]['equation'].render(mapping)
            self.list.add_cid_and_label(cid, [label, expression], editable=False)

        self.list.blockSignals(False)

        self._validate()

    def _validate(self):

        # Construct a list of all labels for the current dataset so that
        # we can check which ones are duplicates
        labels = [c.label for c in self._components_other[self.data]]
        labels.extend([c['label'] for c in self._state[self.data].values()])
        if len(labels) == 0:
            return
        label_count = Counter(labels)

        # It's possible that the duplicates are entirely for components not
        # shown in this editor, so we keep track here of whether an invalid
        # component has been found.
        invalid = False

        if label_count.most_common(1)[0][1] > 1:

            # If we are here, there are duplicates somewhere in the list
            # of components.

            brush_red = QtGui.QBrush(Qt.red)
            brush_black = QtGui.QBrush(Qt.black)

            self.list.blockSignals(True)

            for item in self.list:
                label = item.text(0)
                if label_count[label] > 1:
                    item.setForeground(0, brush_red)
                    invalid = True
                else:
                    item.setForeground(0, brush_black)

            self.list.blockSignals(False)

        if invalid:
            self.ui.label_status.setStyleSheet('color: red')
            self.ui.label_status.setText('Error: some components have duplicate names')
            self.ui.button_ok.setEnabled(False)
            self.ui.combosel_data.setEnabled(False)
        else:
            self.ui.label_status.setStyleSheet('')
            self.ui.label_status.setText('')
            self.ui.button_ok.setEnabled(True)
            self.ui.combosel_data.setEnabled(True)

    def _update_state(self, *args):
        self._components_derived[self.data] = []
        for item in self.list:
            cid = item.data(0, Qt.UserRole)
            self._state[self.data][cid]['label'] = item.text(0)
            self._components_derived[self.data].append(cid)
        self._update_component_lists()

    def _remove_derived_component(self, *args):
        cid = self.list.selected_cid
        if cid is not None:
            self._components_derived[self.data].remove(cid)
            self._state[self.data].pop(cid)
            self._update_component_lists()

    def _add_derived_component(self, *args):

        comp_state = {}
        comp_state['cid'] = ComponentID('')
        comp_state['label'] = ''
        comp_state['equation'] = None

        self._components_derived[self.data].append(comp_state['cid'])
        self._state[self.data][comp_state['cid']] = comp_state

        self._update_component_lists()

        self.list.select_cid(comp_state['cid'])

        result = self._edit_derived_component()

        if not result:  # user cancelled
            self._components_derived[self.data].remove(comp_state['cid'])
            self._state[self.data].pop(comp_state['cid'])
            self._update_component_lists()

    def _edit_derived_component(self, event=None):

        derived_item = self.list.selected_item

        if derived_item is None:
            return False

        derived_cid = self.list.selected_cid

        # Note, we put the pixel/world components last as it's most likely the
        # user wants to use one of the main components.
        mapping = {}
        references = {}
        for cid in (self.data.main_components +
                    self.data.pixel_component_ids +
                    self.data.world_component_ids):
            if cid is not derived_cid:
                mapping[cid] = cid.label
                references[cid.label] = cid

        label = self._state[self.data][derived_cid]['label']

        if self._state[self.data][derived_cid]['equation'] is None:
            equation = None
        else:
            equation = self._state[self.data][derived_cid]['equation'].render(mapping)

        dialog = EquationEditorDialog(label=label, equation=equation, references=references, parent=self)
        dialog.setWindowFlags(self.windowFlags() | Qt.Window)
        dialog.setFocus()
        dialog.raise_()
        dialog.exec_()

        if dialog.final_expression is None:
            return False

        name, equation = dialog.get_final_label_and_parsed_command()
        self._state[self.data][derived_cid]['label'] = name
        self._state[self.data][derived_cid]['equation'] = equation
        derived_item.setText(0, name)

        # Make sure we update the component list here since the equation may
        # have changed and we need to update the preview
        self._update_component_lists()

        return True

    def accept(self):

        for data in self._components_derived:

            cids_derived = self._components_derived[data]
            cids_other = self._components_other[data]
            cids_all = cids_other + cids_derived
            cids_existing = data.components
            components = dict((cid.uuid, cid) for cid in data.components)

            # First deal with renaming of components
            for cid_new in cids_derived:
                label = self._state[data][cid_new]['label']
                if label != cid_new.label:
                    cid_new.label = label

            # Second deal with the removal of components
            for cid_old in cids_existing:
                if not any(cid_old is cid_new for cid_new in cids_all):
                    data.remove_component(cid_old)

            # Third, update/add arithmetic expressions as needed
            for cid_new in cids_derived:
                if any(cid_new is cid_old for cid_old in cids_existing):
                    comp = data.get_component(cid_new)
                    if comp.link._parsed._cmd != self._state[data][cid_new]['equation']._cmd:
                        comp.link._parsed._cmd = self._state[data][cid_new]['equation']._cmd
                        comp.link._parsed._references = components
                        if data.hub:
                            msg = NumericalDataChangedMessage(data)
                            data.hub.broadcast(msg)
                else:
                    pc = ParsedCommand(self._state[data][cid_new]['equation']._cmd, components)
                    link = ParsedComponentLink(cid_new, pc)
                    data.add_component_link(link)

            # Findally, reorder components as needed
            data.reorder_components(cids_all)

        super(ArithmeticEditorWidget, self).accept()


if __name__ == "__main__":  # pragma: nocover

    from glue.utils.qt import get_qapp
    app = get_qapp()

    import numpy as np

    from glue.core.data import Data
    from glue.core.data_collection import DataCollection

    x = np.random.random((5, 5))
    y = x * 3
    dc = DataCollection()
    dc.append(Data(label='test1', x=x, y=y))
    dc.append(Data(label='test2', a=x, b=y))

    widget = ArithmeticEditorWidget(dc)
    widget.exec_()



from __future__ import absolute_import, division, print_function

from .astropy_table import *  # noqa
from .dendrogram import *  # noqa
from .excel import *  # noqa
from .fits import *  # noqa
from .hdf5 import *  # noqa
from .helpers import *  # noqa
from .image import *  # noqa
from .numpy import *  # noqa
from .pandas import *  # noqa
from .tables import *  # noqa


# from .deprecated import *



from __future__ import absolute_import, division, print_function

import os
import warnings
import traceback
from functools import wraps

from glue.core.data import Subset
from glue.external.six import string_types
from glue.core.session import Session
from glue.core.hub import HubListener
from glue.core import BaseData
from glue.core.data_factories import load_data
from glue.core.data_collection import DataCollection
from glue.config import settings


__all__ = ['Application']


def catch_error(msg):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                m = "%s\n%s" % (msg, str(e))
                detail = str(traceback.format_exc())
                self = args[0]
                self.report_error(m, detail)
        return wrapper
    return decorator


def as_flat_data_list(data):
    datasets = []
    if isinstance(data, BaseData):
        datasets.append(data)
    else:
        for d in data:
            datasets.extend(as_flat_data_list(d))
    return datasets


class Application(HubListener):

    def __init__(self, data_collection=None, session=None):
        if session is not None:
            self._session = session
            session.application = self
            self._data = session.data_collection
        else:
            self._data = data_collection or DataCollection()
            self._session = Session(data_collection=self._data,
                                    application=self)

        self._hub = self._session.hub
        self._cmds = self._session.command_stack
        self._cmds.add_callback(self._update_undo_redo_enabled)

        self._settings = {}
        for key, value, validator in settings:
            self._settings[key] = [value, validator]

    @property
    def session(self):
        return self._session

    @property
    def data_collection(self):
        return self.session.data_collection

    def new_data_viewer(self, viewer_class, data=None, state=None):
        """
        Create a new data viewer, add it to the UI,
        and populate with data
        """
        if viewer_class is None:
            return

        if state is not None:
            c = viewer_class(self._session, state=state)
        else:
            c = viewer_class(self._session)
        c.register_to_hub(self._session.hub)

        if data is not None:
            if isinstance(data, BaseData):
                result = c.add_data(data)
            elif isinstance(data, Subset):
                result = c.add_subset(data)
            if not result:
                c.close(warn=False)
                return

        self.add_widget(c)
        return c

    @catch_error("Failed to save session")
    def save_session(self, path, include_data=False, absolute_paths=True):
        """ Save the data collection and hub to file.

        Can be restored via restore_session

        Note: Saving of client is not currently supported. Thus,
        restoring this session will lose all current viz windows
        """

        from glue.core.state import GlueSerializer
        gs = GlueSerializer(self,
                            include_data=include_data,
                            absolute_paths=absolute_paths)

        # In case relative paths are needed in the session file, we do the
        # serialization while setting the current directory to the directory
        # in which the session file will be saved so that the relative paths
        # are relative to the session file, not the current working directory.
        start_dir = os.path.abspath('.')
        session_dir = os.path.dirname(path) or '.'

        try:
            os.chdir(session_dir)
            state = gs.dumps(indent=2)
        finally:
            os.chdir(start_dir)

        with open(path, 'w') as out:
            out.write(state)

    @staticmethod
    def restore_session(path):
        """
        Reload a previously-saved session

        Parameters
        ----------
        path : str
            Path to the file to load

        Returns
        -------
        app : :class:`Application`
            The loaded application
        """
        from glue.core.state import GlueUnSerializer

        # In case relative paths are needed in the session file, we do the
        # loading while setting the current directory to the directory
        # in which the session file is so that relative paths are interpreted
        # as relative to the session file.
        start_dir = os.path.abspath('.')
        session_dir = os.path.dirname(path) or '.'
        session_file = os.path.basename(path)

        try:
            os.chdir(session_dir)
            with open(session_file) as infile:
                state = GlueUnSerializer.load(infile)
            return state.object('__main__')
        finally:
            os.chdir(start_dir)

    def get_setting(self, key):
        """
        Fetch the value of an application setting
        """
        return self._settings[key][0]

    def set_setting(self, key, value):
        """
        Set the value of an application setting

        Raises a KeyError if the setting does not exist
        Raises a ValueError if the value is invalid
        """
        validator = self._settings[key][1]
        self._settings[key][0] = validator(value)

    @property
    def settings(self):
        """Iterate over settings"""
        for key, (value, _) in self._settings.items():
            yield key, value

    @catch_error("Could not load data")
    def load_data(self, paths, auto_merge=False, **kwargs):
        """
        Given a path to a file, load the file as a Data object and add it to
        the current session.

        This returns the added `Data` object.
        """

        if isinstance(paths, string_types):
            paths = [paths]

        datasets = []
        for path in paths:
            result = load_data(path)
            if isinstance(result, BaseData):
                datasets.append(result)
            else:
                datasets.extend(result)

        self.add_datasets(datasets, auto_merge=auto_merge, **kwargs)

        if len(datasets) == 1:
            return datasets[0]
        else:
            return datasets

    @catch_error("Could not add data")
    def add_data(self, *args, **kwargs):
        """
        Add data to the session.

        Positional arguments are interpreted using the data factories, while
        keyword arguments are interpreted using the same infrastructure as the
        `qglue` command.

        This returns a list of added `Data` objects.
        """

        datasets = []

        for path in args:
            if isinstance(path, BaseData):
                datasets.append(path)
            else:
                datasets.append(load_data(path))

        links = kwargs.pop('links', None)

        from glue.qglue import parse_data, parse_links

        for label, data in kwargs.items():
            datasets.extend(parse_data(data, label))

        self.add_datasets(datasets)

        if links is not None:
            self.data_collection.add_link(parse_links(self.data_collection, links))

        return datasets

    def report_error(self, message, detail):
        """ Report an error message to the user.
        Must be implemented in a subclass

        Parameters
        ----------
        message : str
            The message to display
        detail : str
            Longer context about the error
        """
        raise Exception(message + "(" + detail + ")")

    def do(self, command):
        return self._cmds.do(command)

    def undo(self):
        try:
            self._cmds.undo()
        except RuntimeError:
            pass

    def redo(self):
        try:
            self._cmds.redo()
        except RuntimeError:
            pass

    def _update_undo_redo_enabled(self):
        raise NotImplementedError()

    def add_datasets(self, *args, **kwargs):
        """
        Utility method to interactively add datasets to the data_collection.

        Parameters
        ----------
        datasets : :class:`~glue.core.data.Data` or list of Data
            One or more :class:`~glue.core.data.Data` instances.

        Adds datasets to the collection in the application.
        """

        if isinstance(args[0], DataCollection):
            warnings.warn("Calling add_datasets with an explicit data "
                          "collection is now deprecated. You should now call "
                          "app.add_datasets(datasets).", UserWarning)
            data_collection, datasets = args
        else:
            data_collection = self.data_collection
            datasets = args[0]

        if "skip_merge" in kwargs:
            warnings.warn("The skip_merge= argument now no longer has any "
                          "effect and is deprecated, since no merging is done "
                          "by default", UserWarning)

        auto_merge = kwargs.pop('auto_merge', False)

        datasets = as_flat_data_list(datasets)
        data_collection.extend(datasets)

        # We now automatically merge the datasets if requested. However, we only
        # do this if all the datasets have the same shape to avoid confusion.
        if auto_merge and len(datasets) > 1:

            reference_shape = datasets[0].shape

            if all([data.shape == reference_shape for data in datasets[1:]]):
                suggested_label = data_collection.suggest_merge_label(*datasets)
                return data_collection.merge(*datasets, label=suggested_label)
            else:
                raise ValueError("Can only auto-merge datasets if they all have "
                                 " the same shape")

        else:

            return datasets

    @staticmethod
    def _choose_merge(data, other, suggested_label):
        """
        Present an interface to the user for approving or rejecting
        a proposed data merger. Returns a list of datasets from other
        that the user has approved to merge with data
        """
        raise NotImplementedError

    @property
    def viewers(self):
        """
        Return a tuple of tuples of viewers currently open.
        """
        return []

    def set_data_color(self, color, alpha):
        """
        Reset all the data colors to that specified.
        """
        for data in self.data_collection:
            data.style.color = color
            data.style.alpha = alpha

    def __gluestate__(self, context):
        viewers = [list(map(context.id, tab)) for tab in self.viewers]
        data = self.session.data_collection
        from glue.main import _loaded_plugins
        return dict(session=context.id(self.session), viewers=viewers,
                    data=context.id(data), plugins=_loaded_plugins)

    @classmethod
    def __setgluestate__(cls, rec, context):
        self = cls(data_collection=context.object(rec['data']))
        # manually register the newly-created session, which
        # the viewers need
        context.register_object(rec['session'], self.session)
        for i, tab in enumerate(rec['viewers']):
            if self.tab(i) is None:
                self.new_tab()
            for v in tab:
                viewer = context.object(v)
                self.add_widget(viewer, tab=i, hold_position=True)
        return self



from distutils.version import LooseVersion

from astropy import __version__

# The autolinking functionality relies on the new WCS API present in
# Astropy 3.1 and later.
ASTROPY_GE_31 = LooseVersion(__version__) >= '3.1'


def setup():

    if not ASTROPY_GE_31:
        return

    from . import wcs_autolinking  # noqa



from __future__ import absolute_import, division, print_function

import numpy as np
import matplotlib.pyplot as plt
from astropy.utils import NumpyRNGContext

from glue.external.six import PY2
from glue.core import Data, DataCollection
from glue.app.qt.application import GlueApplication
from glue.viewers.image.qt import ImageViewer
from glue.viewers.matplotlib.qt.tests.test_python_export import BaseTestExportPython


class TestExportPython(BaseTestExportPython):

    def setup_method(self, method):

        with NumpyRNGContext(12345):
            self.data = Data(cube=np.random.random((30, 50, 20)))
        self.data_collection = DataCollection([self.data])
        self.app = GlueApplication(self.data_collection)
        self.viewer = self.app.new_data_viewer(ImageViewer)
        self.viewer.add_data(self.data)
        # FIXME: On some platforms, using an integer label size
        # causes some of the labels to be non-deterministically
        # shifted by one pixel, so we pick a non-round font size
        # to avoid this.
        self.viewer.state.x_ticklabel_size = 8.21334111
        self.viewer.state.y_ticklabel_size = 8.21334111

    def teardown_method(self, method):
        self.viewer.close()
        self.viewer = None
        self.app.close()
        self.app = None

    def assert_same(self, tmpdir):
        # Due to numerical issues, the tests are not pixel perfect on Python 2
        BaseTestExportPython.assert_same(self, tmpdir, tol=5 if PY2 else 0.1)

    def test_simple(self, tmpdir):
        self.assert_same(tmpdir)

    def test_simple_att(self, tmpdir):
        self.viewer.state.x_att = self.data.pixel_component_ids[1]
        self.viewer.state.y_att = self.data.pixel_component_ids[0]
        self.assert_same(tmpdir)

    def test_simple_visual(self, tmpdir):
        self.viewer.state.layers[0].cmap = plt.cm.RdBu
        self.viewer.state.layers[0].v_min = 0.2
        self.viewer.state.layers[0].v_max = 0.8
        self.viewer.state.layers[0].stretch = 'sqrt'
        self.viewer.state.layers[0].stretch = 'sqrt'
        self.viewer.state.layers[0].contrast = 0.9
        self.viewer.state.layers[0].bias = 0.6
        self.assert_same(tmpdir)

    def test_slice(self, tmpdir):
        self.viewer.state.x_att = self.data.pixel_component_ids[1]
        self.viewer.state.y_att = self.data.pixel_component_ids[0]
        self.viewer.state.slices = (2, 3, 4)
        self.assert_same(tmpdir)

    def test_aspect(self, tmpdir):
        self.viewer.state.aspect = 'auto'
        self.assert_same(tmpdir)

    def test_subset(self, tmpdir):
        self.data_collection.new_subset_group('mysubset', self.data.id['cube'] > 0.5)
        self.assert_same(tmpdir)

    def test_subset_slice(self, tmpdir):
        self.data_collection.new_subset_group('mysubset', self.data.id['cube'] > 0.5)
        self.test_slice(tmpdir)

    def test_subset_transposed(self, tmpdir):
        self.viewer.state.x_att = self.data.pixel_component_ids[0]
        self.viewer.state.y_att = self.data.pixel_component_ids[1]
        self.data_collection.new_subset_group('mysubset', self.data.id['cube'] > 0.5)
        self.assert_same(tmpdir)



from __future__ import absolute_import, division, print_function

import uuid
import numbers
import operator

import numpy as np

from glue.external import six
from glue.external.six import PY3
from glue.core.roi import (PolygonalROI, CategoricalROI, RangeROI, XRangeROI,
                           YRangeROI, RectangularROI)
from glue.core.contracts import contract
from glue.core.util import split_component_view
from glue.core.registry import Registry
from glue.core.exceptions import IncompatibleAttribute
from glue.core.message import SubsetDeleteMessage, SubsetUpdateMessage
from glue.core.decorators import memoize
from glue.core.visual import VisualAttributes
from glue.config import settings
from glue.utils import (view_shape, broadcast_to, floodfill, combine_slices,
                        polygon_line_intersections, categorical_ndarray)


__all__ = ['Subset', 'SubsetState', 'RoiSubsetState', 'CategoricalROISubsetState',
           'RangeSubsetState', 'MultiRangeSubsetState', 'CompositeSubsetState',
           'OrState', 'AndState', 'XorState', 'InvertState', 'MaskSubsetState', 'CategorySubsetState',
           'ElementSubsetState', 'InequalitySubsetState', 'combine_multiple',
           'CategoricalMultiRangeSubsetState', 'CategoricalROISubsetState2D',
           'SliceSubsetState', 'roi_to_subset_state', 'MultiOrState']


OPSYM = {operator.ge: '>=', operator.gt: '>',
         operator.le: '<=', operator.lt: '<',
         operator.and_: '&', operator.or_: '|',
         operator.xor: '^', operator.eq: '==',
         operator.ne: '!='}
SYMOP = dict((v, k) for k, v in OPSYM.items())


class Subset(object):

    """Base class to handle subsets of data.

    These objects both describe subsets of a dataset, and relay any
    state changes to the hub that their parent data are assigned to.

    This base class only directly implements the logic that relays
    state changes back to the hub. Subclasses implement the actual
    description and manipulation of data subsets

    :param data:
        The dataset that this subset describes
    :type data: :class:`~glue.core.data.Data`
    """

    @contract(data='isinstance(Data)|None',
              color='color',
              alpha=float,
              label='string|None')
    def __init__(self, data, color=settings.SUBSET_COLORS[0], alpha=0.5, label=None):
        """ Create a new subset object.

        Note: the preferred way for creating subsets is
        via DataCollection.new_subset_group. Manually-instantiated
        subsets will probably *not* be represented properly by the UI
        """

        self._broadcasting = False  # must be first def

        self.data = data
        self.label = label  # trigger disambiguation

        self.subset_state = SubsetState()  # calls proper setter method

        self.style = VisualAttributes(parent=self)
        self.style.markersize *= 2.5
        self.style.linewidth *= 2.5
        self.style.color = color
        self.style.alpha = alpha

        # We assign a UUID which can then be used for example in equations
        # for derived components - the idea is that this doesn't change over
        # the life cycle of glue, so it is a more reliable way to refer to
        # components in strings than using labels
        self._uuid = str(uuid.uuid4())

    @property
    def uuid(self):
        return self._uuid

    @property
    def subset_state(self):
        return self._subset_state

    @subset_state.setter
    def subset_state(self, state):
        if isinstance(state, np.ndarray):
            if self.data.shape != state.shape:
                raise ValueError("Shape of mask doesn't match shape of data")
            cids = self.data.pixel_component_ids
            state = MaskSubsetState(state, cids)
        if not isinstance(state, SubsetState):
            raise TypeError("State must be a SubsetState instance or array")
        self._subset_state = state

    @property
    def style(self):
        return self._style

    @style.setter
    @contract(value=VisualAttributes)
    def style(self, value):
        value.parent = self
        self._style = value

    @property
    def label(self):
        """ Convenience access to subset's label """
        return self._label

    @label.setter
    def label(self, value):
        """Set the subset's label

        Subset labels within a data object must be unique. The input
        will be auto-disambiguated if necessary
        """
        value = Registry().register(self, value, group=self.data)
        self._label = value

    @property
    def attributes(self):
        """
        Returns a tuple of the ComponentIDs that this subset
        depends upon
        """
        return self.subset_state.attributes

    def register(self):
        """ Register a subset to its data, and start broadcasting
        state changes

        """
        self.data.add_subset(self)
        self.do_broadcast(True)

    @contract(returns='array[N]')
    def to_index_list(self):
        """
        Convert the current subset to a list of indices. These index
        the elements in the (flattened) data object that belong to the subset.

        If x is the numpy array corresponding to some component.data,
        the two following statements are equivalent::

           x.flat[subset.to_index_list()]
           x[subset.to_mask()]

        Returns:

           A numpy array, giving the indices of elements in the data that
           belong to this subset.

        Raises:

           IncompatibleDataException: if an index list cannot be created
           for the requested data set.

        """
        return self.subset_state.to_index_list(self.data)

    @contract(view='array_view', returns='array')
    def to_mask(self, view=None):
        """
        Convert the current subset to a mask.

        :param view: An optional view into the dataset (e.g. a slice)
                     If present, the mask will pertain to the view and not the
                     entire dataset.

           A boolean numpy array, the same shape as the data, that
           defines whether each element belongs to the subset.

        """
        return self.data.get_mask(self.subset_state, view=view)

    @contract(value=bool)
    def do_broadcast(self, value):
        """
        Set whether state changes to the subset are relayed to a hub.

        It can be useful to turn off broadcasting, when modifying the
        subset in ways that don't impact any of the clients.

        Attributes:
        value: Whether the subset should broadcast state changes (True/False)

        """
        object.__setattr__(self, '_broadcasting', value)

    @contract(attribute='string')
    def broadcast(self, attribute):
        """
        Explicitly broadcast a SubsetUpdateMessage to the hub

        :param attribute:
                   The name of the attribute (if any) that should be
                   broadcast as updated.
        :type attribute: ``str``

        """

        if not hasattr(self, 'data') or not hasattr(self.data, 'hub'):
            return

        if self._broadcasting and self.data.hub:
            msg = SubsetUpdateMessage(self, attribute=attribute)
            self.data.hub.broadcast(msg)

    def delete(self):
        """Broadcast a SubsetDeleteMessage to the hub, and stop broadcasting

        Also removes subset reference from parent data's subsets list
        """

        dobroad = self._broadcasting and self.data is not None and \
            self.data.hub is not None

        self.do_broadcast(False)

        if self.data is not None and self in self.data.subsets:
            self.data._subsets.remove(self)

        if dobroad:
            msg = SubsetDeleteMessage(self)
            self.data.hub.broadcast(msg)

        Registry().unregister(self, group=self.data)

    @contract(file_name='string')
    def write_mask(self, file_name, format="fits"):
        """ Write a subset mask out to file

        :param file_name: name of file to write to
        :param format:
           Name of format to write to. Currently, only "fits" is
           supported

        """
        mask = np.short(self.to_mask())
        if format == 'fits':
            from astropy.io import fits
            try:
                fits.writeto(file_name, mask, overwrite=True)
            except TypeError:
                fits.writeto(file_name, mask, clobber=True)
        else:
            raise AttributeError("format not supported: %s" % format)

    @contract(file_name='string')
    def read_mask(self, file_name):
        try:
            from astropy.io import fits
            with fits.open(file_name) as hdulist:
                mask = hdulist[0].data
        except IOError:
            raise IOError("Could not read %s (not a fits file?)" % file_name)
        ind = np.where(mask.flat)[0]
        state = ElementSubsetState(indices=ind)
        self.subset_state = state

    def __del__(self):
        try:
            self.delete()
        except Exception:
            pass

    def __setattr__(self, attribute, value):
        had_attribute = hasattr(self, attribute)
        before = getattr(self, attribute, None)
        object.__setattr__(self, attribute, value)
        if not attribute.startswith('_') and (not had_attribute or np.any(before != value)):
            self.broadcast(attribute)

    def __getitem__(self, view):
        """ Retrieve the elements from a data view within the subset

        :param view: View of the data. See data.__getitem__ for detils
        """
        c, v = split_component_view(view)
        ma = self.to_mask(v)
        return self.data[view][ma]

    @contract(other_subset='isinstance(Subset)')
    def paste(self, other_subset):
        """paste subset state from other_subset onto self """
        state = other_subset.subset_state.copy()
        self.subset_state = state

    def __str__(self):
        dlabel = "(no data)"
        if self.data is not None:
            dlabel = "(data: %s)" % self.data.label
        slabel = "Subset: (no label)"
        if self.label:
            slabel = "Subset: %s" % self.label
        return "%s %s" % (slabel, dlabel)

    def __repr__(self):
        return self.__str__()

    @contract(other='isinstance(Subset)', returns='isinstance(Subset)')
    def __or__(self, other):
        return _combine([self, other], operator.or_)

    @contract(other='isinstance(Subset)', returns='isinstance(Subset)')
    def __and__(self, other):
        return _combine([self, other], operator.and_)

    @contract(returns='isinstance(Subset)')
    def __invert__(self):
        return _combine([self], operator.invert)

    @contract(other='isinstance(Subset)', returns='isinstance(Subset)')
    def __xor__(self, other):
        return _combine([self, other], operator.xor)

    def __eq__(self, other):
        if not isinstance(other, Subset):
            return False
        # XXX need to add equality specification for subset states
        if self is other:
            return True
        return (self.subset_state == other.subset_state and
                self.style == other.style)

    def state_as_mask(self):
        """
        Convert the current SubsetState to a MaskSubsetState
        """
        try:
            m = self.to_mask()
        except IncompatibleAttribute:
            m = np.zeros(self.data.shape, dtype=np.bool)
        cids = self.data.pixel_component_ids
        return MaskSubsetState(m, cids)

    # In Python 2 we need to do this explicitly
    def __ne__(self, other):
        return not self.__eq__(other)

    # In Python 3, if __eq__ is defined, then __hash__ has to be re-defined
    if PY3:
        __hash__ = object.__hash__

    # Provide convenient access to Data methods/properties that make sense
    # here too.

    def component_ids(self):
        return self.data.component_ids()

    @property
    def components(self):
        return self.data.components

    @property
    def derived_components(self):
        return self.data.derived_components

    @property
    def main_components(self):
        return self.data.main_components

    @property
    def pixel_component_ids(self):
        return self.data.pixel_component_ids

    @property
    def world_component_ids(self):
        return self.data.world_component_ids

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def shape(self):
        return self.data.shape

    @property
    def size(self):
        return self.data.size

    @property
    def hub(self):
        return self.data.hub

    # DEPRECATED (warnings raised in Data)

    @property
    def primary_components(self):
        return self.data.primary_components

    @property
    def visible_components(self):
        return self.data.visible_components


class SubsetState(object):
    """
    The base class for all subset states.

    This defaults to an empty subset.
    """

    def __init__(self):
        pass

    @property
    def attributes(self):
        """
        The attributes that the subset state depends on.
        """
        return tuple()

    @property
    def subset_state(self):  # convenience method, mimic interface of Subset
        return self

    @contract(data='isinstance(Data)')
    def to_index_list(self, data):
        return np.where(data.get_mask(self.subset_state).flat)[0]

    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        """
        Compute the mask for this subset state.

        Parameters
        ----------
        data : :class:`~glue.core.data.Data`
            The dataset to compute the mask for.
        view
            Any object that returns a valid view for a Numpy array.
        """
        shp = view_shape(data.shape, view)
        return broadcast_to(False, shp)

    @contract(returns='isinstance(SubsetState)')
    def copy(self):
        """
        Return a copy of the subset state.
        """
        return SubsetState()

    @contract(other_state='isinstance(SubsetState)',
              returns='isinstance(SubsetState)')
    def __or__(self, other_state):
        return OrState(self, other_state)

    @contract(other_state='isinstance(SubsetState)',
              returns='isinstance(SubsetState)')
    def __and__(self, other_state):
        return AndState(self, other_state)

    @contract(returns='isinstance(SubsetState)')
    def __invert__(self):
        return InvertState(self)

    @contract(other_state='isinstance(SubsetState)',
              returns='isinstance(SubsetState)')
    def __xor__(self, other_state):
        return XorState(self, other_state)


class RoiSubsetState(SubsetState):
    """
    A subset defined as the set of points in two dimensions that lie inside
    a region of interest (ROI).

    The two dimensions are defined as two numerical data attributes.

    Parameters
    ----------
    xatt : :class:`~glue.core.component_id.ComponentID`
        The data attribute on the x axis.
    yatt : :class:`~glue.core.component_id.ComponentID`
        The data attribute on the y axis.
    roi : :class:`~glue.core.roi.Roi`
        The region of interest.
    """

    @contract(xatt='isinstance(ComponentID)', yatt='isinstance(ComponentID)')
    def __init__(self, xatt=None, yatt=None, roi=None):
        super(RoiSubsetState, self).__init__()
        self._xatt = xatt
        self._yatt = yatt
        self._roi = roi

    @property
    def xatt(self):
        """
        The data attribute on the x axis.
        """
        return self._xatt

    @xatt.setter
    def xatt(self, value):
        self._xatt = value

    @property
    def yatt(self):
        """
        The data attribute on the y axis.
        """
        return self._yatt

    @yatt.setter
    def yatt(self, value):
        self._yatt = value

    @property
    def roi(self):
        """
        The region of interest.
        """
        return self._roi

    @roi.setter
    def roi(self, value):
        self._roi = value

    @property
    def attributes(self):
        return (self.xatt, self.yatt)

    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):

        # TODO: make sure that pixel components don't actually take up much
        #       memory and are just views

        x = data[self.xatt, view]
        y = data[self.yatt, view]

        if (x.ndim == data.ndim and
            self.xatt in data.pixel_component_ids and
                self.yatt in data.pixel_component_ids):

            # This is a special case - the ROI is defined in pixel space, so we
            # can apply it to a single slice and then broadcast it to all other
            # dimensions. We start off by extracting a slice which takes only
            # the first elements of all dimensions except the attributes in
            # question, for which we take all the elements. We need to preserve
            # the dimensionality of the array, hence the use of slice(0, 1).
            # Note that we can only do this if the view (if present) preserved
            # the dimensionality, which is why we checked that x.ndim == data.ndim

            subset = []
            for i in range(data.ndim):
                if i == self.xatt.axis or i == self.yatt.axis:
                    subset.append(slice(None))
                else:
                    subset.append(slice(0, 1))

            x_slice = x[tuple(subset)]
            y_slice = y[tuple(subset)]

            if self.roi.defined():
                result = self.roi.contains(x_slice, y_slice)
            else:
                result = np.zeros(x_slice.shape, dtype=bool)

            result = broadcast_to(result, x.shape)

        else:

            if self.roi.defined():
                result = self.roi.contains(x, y)
            else:
                result = np.zeros(x.shape, dtype=bool)

        if result.shape != x.shape:
            raise ValueError("Unexpected error: boolean mask has incorrect dimensions")

        return result

    def copy(self):
        result = RoiSubsetState()
        result.xatt = self.xatt
        result.yatt = self.yatt
        result.roi = self.roi
        return result


class CategoricalROISubsetState(SubsetState):
    """
    A subset defined as the set of values for a categorical data attribute that
    fall inside a categorical region of interest (ROI).

    Parameters
    ----------
    att : :class:`~glue.core.component_id.ComponentID`
        The categorical data attribute used for the subset.
    roi : :class:`~glue.core.roi.CategoricalROI`
        The categorical region of interest.
    """

    def __init__(self, att=None, roi=None):
        super(CategoricalROISubsetState, self).__init__()
        self._att = att
        self._roi = roi

    @property
    def att(self):
        """
        The categorical data attribute used for the subset.
        """
        return self._att

    @att.setter
    def att(self, value):
        self._att = value

    @property
    def roi(self):
        """
        The categorical region of interest.
        """
        return self._roi

    @roi.setter
    def roi(self, value):
        self._roi = value

    @property
    def attributes(self):
        return self.att,

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        x = data[self.att, view]
        result = self.roi.contains(x, None)
        assert x.shape == result.shape
        return result.ravel()

    def copy(self):
        result = CategoricalROISubsetState()
        result.att = self.att
        result.roi = self.roi
        return result

    @staticmethod
    def from_range(categories, att, lo, hi):

        roi = CategoricalROI.from_range(categories, lo, hi)
        subset = CategoricalROISubsetState(roi=roi, att=att)
        return subset

    def __gluestate__(self, context):
        return dict(att=context.id(self.att),
                    roi=context.id(self.roi))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(att=context.object(rec['att']), roi=context.object(rec['roi']))


class RangeSubsetState(SubsetState):
    """
    A subset defined as the set of values inside a range.

    The range is defined as being inclusive (that is, values equal to the lower
    or upper bounds are considered to be inside the subset).

    Parameters
    ----------
    lo : `float`
        The lower limit of the range.
    hi : `float`
        The upper limit of the range.
    att : :class:`~glue.core.component_id.ComponentID`
        The attribute being used for the subset.
    """

    def __init__(self, lo, hi, att=None):
        super(RangeSubsetState, self).__init__()
        self._lo = lo
        self._hi = hi
        self._att = att

    @property
    def lo(self):
        """
        The lower limit of the range.
        """
        return self._lo

    @lo.setter
    def lo(self, value):
        self._lo = value

    @property
    def hi(self):
        """
        The upper limit of the range.
        """
        return self._hi

    @hi.setter
    def hi(self, value):
        self._hi = value

    @property
    def att(self):
        """
        The attribute being used for the subset.
        """
        return self._att

    @att.setter
    def att(self, value):
        self._att = value

    @property
    def attributes(self):
        return (self.att,)

    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        x = data[self.att, view]
        result = (x >= self.lo) & (x <= self.hi)
        return result

    def copy(self):
        return RangeSubsetState(self.lo, self.hi, self.att)


class MultiRangeSubsetState(SubsetState):
    """
    A subset state defined by multiple discontinuous ranges

    The ranges are defined as being inclusive (that is, values equal to the
    lower or upper bounds are considered to be inside the subset).

    Parameters
    ----------
    pairs : list
        A list of (lo, hi) tuples.
    att : :class:`~glue.core.component_id.ComponentID`
        The attribute being used for the subset.
    """

    def __init__(self, pairs, att=None):
        super(MultiRangeSubsetState, self).__init__()
        self._pairs = pairs
        self._att = att

    @property
    def pairs(self):
        """
        A list of (lo, hi) tuples.
        """
        return self._pairs

    @pairs.setter
    def pairs(self, value):
        self._pairs = value

    @property
    def att(self):
        """
        The attribute being used for the subset.
        """
        return self._att

    @att.setter
    def att(self, value):
        self._att = value

    @property
    def attributes(self):
        return (self.att,)

    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        x = data[self.att, view]
        result = np.zeros_like(x, dtype=bool)
        for lo, hi in self.pairs:
            result |= (x >= lo) & (x <= hi)
        return result

    def copy(self):
        return MultiRangeSubsetState(self.pairs, self.att)


class CategoricalROISubsetState2D(SubsetState):
    """
    A subset defined as the set of values for two categorical data attributes
    that fall inside a categorical region of interest (ROI).

    Parameters
    ----------
    categories : dict
        A dictionary containing for each label of one categorical component an
        iterable of labels for the other categorical component (using sets will
        provide the best performance)
    att1 : :class:`~glue.core.component_id.ComponentID`
        The component ID matching the keys of the ``categories`` dictionary
    att2 : :class:`~glue.core.component_id.ComponentID`
        The component ID matching the values of the ``categories`` dictionary
    """

    def __init__(self, categories, att1, att2):
        self._categories = categories
        self._att1 = att1
        self._att2 = att2

    @property
    def categories(self):
        """
        A dictionary containing for each label of one categorical component an
        iterable of labels for the other categorical component.
        """
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

    @property
    def att1(self):
        """
        The component ID matching the keys of the ``categories`` dictionary
        """
        return self._att1

    @att1.setter
    def att1(self, value):
        self._att1 = value

    @property
    def att2(self):
        """
        The component ID matching the values of the ``categories`` dictionary
        """
        return self._att2

    @att2.setter
    def att2(self, value):
        self._att2 = value

    @property
    def attributes(self):
        return (self.att1, self.att2)

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):

        # Extract categories and numerical values
        labels1 = data[self.att1, view]
        labels2 = data[self.att2, view]

        # Initialize empty mask
        mask = np.zeros(labels1.shape, dtype=bool)

        # A loop over all values here is actually reasonably efficient compared
        # to alternatives. Any improved implementation, even vectorized, should
        # ensure that it is more efficient for large numbers of categories and
        # values.
        for i in range(len(labels1)):
            if labels1[i] in self.categories:
                if labels2[i] in self.categories[labels1[i]]:
                    mask[i] = True

        return mask

    def copy(self):
        result = CategoricalROISubsetState2D(self.categories,
                                             self.att1, self.att2)
        return result

    def __gluestate__(self, context):
        return dict(categories=self.categories,
                    att1=context.id(self.att1),
                    att2=context.id(self.att2))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(categories=rec['categories'],
                   att1=context.object(rec['att1']),
                   att2=context.object(rec['att2']))


class CategoricalMultiRangeSubsetState(SubsetState):
    """
    A subset state defined by two attributes where one attribute is categorical
    and the other is numerical, and where for each category, there are multiple
    possible subset ranges.

    Parameters
    ----------
    ranges : dict
        A dictionary containing for each category (key), a list of tuples
        giving the ranges of values for the numerical attribute.
    cat_att : :class:`~glue.core.component_id.ComponentID`
        The component ID for the categorical attribute.
    num_att : :class:`~glue.core.component_id.ComponentID`
        The component ID for the numerical attribute.
    """

    def __init__(self, ranges, cat_att, num_att):
        self.ranges = ranges
        self.cat_att = cat_att
        self.num_att = num_att

    @property
    def ranges(self):
        """
        A dictionary containing for each category (key), a list of tuples
        giving the ranges of values for the numerical attribute.
        """
        return self._ranges

    @ranges.setter
    def ranges(self, value):
        self._ranges = value

    @property
    def cat_att(self):
        """
        The component ID for the categorical attribute.
        """
        return self._cat_att

    @cat_att.setter
    def cat_att(self, value):
        self._cat_att = value

    @property
    def num_att(self):
        """
        The component ID for the numerical attribute.
        """
        return self._num_att

    @num_att.setter
    def num_att(self, value):
        self._num_att = value

    @property
    def attributes(self):
        return (self.cat_att, self._num_att)

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):

        # Extract categories and numerical values
        labels = data[self.cat_att]
        values = data[self.num_att]

        if view is not None:
            labels = labels[view]
            values = values[view]

        # Initialize empty mask
        mask = np.zeros(values.shape, dtype=bool)

        # A loop over all values here is actually reasonably efficient compared
        # to alternatives. Any improved implementation, even vectorized, should
        # ensure that it is more efficient for large numbers of categories and
        # values. For example, using 10000 categories and 1000000 data points
        # takes 1.2 seconds on a laptop.
        for i in range(len(values)):
            if labels[i] in self.ranges:
                for lo, hi in self.ranges[labels[i]]:
                    if values[i] >= lo and values[i] <= hi:
                        mask[i] = True
                        break

        return mask

    def copy(self):
        result = CategoricalMultiRangeSubsetState(self.ranges,
                                                  self.cat_att,
                                                  self.num_att)
        return result

    def __gluestate__(self, context):
        return dict(ranges=self.ranges,
                    cat_att=context.id(self.cat_att),
                    num_att=context.id(self.num_att))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(ranges=rec['ranges'],
                   cat_att=context.object(rec['cat_att']),
                   num_att=context.object(rec['num_att']))


class CompositeSubsetState(SubsetState):
    """
    The base class for combinations of subset states.
    """

    op = None

    def __init__(self, state1, state2=None):
        super(CompositeSubsetState, self).__init__()
        self.state1 = state1.copy()
        if state2:
            state2 = state2.copy()
        self.state2 = state2

    def copy(self):
        return type(self)(self.state1, self.state2)

    @property
    def attributes(self):
        att = self.state1.attributes
        if self.state2 is not None:
            att += self.state2.attributes
        return tuple(sorted(set(att)))

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        return self.op(self.state1.to_mask(data, view),
                       self.state2.to_mask(data, view))

    def __str__(self):
        sym = OPSYM.get(self.op, self.op)
        return "(%s %s %s)" % (self.state1, sym, self.state2)


class OrState(CompositeSubsetState):
    """
    An 'or' logical combination of subset states.

    The two states can be accessed using the attributes ``state1`` and
    ``state2``.
    """
    op = operator.or_


class AndState(CompositeSubsetState):
    """
    An 'and' logical combination of subset states.

    The two states can be accessed using the attributes ``state1`` and
    ``state2``.
    """
    op = operator.and_


class XorState(CompositeSubsetState):
    """
    An 'exclusive or' logical combination of subset states.

    The two states can be accessed using the attributes ``state1`` and
    ``state2``.
    """
    op = operator.xor


class InvertState(CompositeSubsetState):
    """
    A inverted subset state.

    Values inside the original subset are now considered outside, and
    vice-versa. The original subset state can be accessed using the attribute
    ``state1``.
    """

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        return ~self.state1.to_mask(data, view)

    def __str__(self):
        return "(~%s)" % self.state1


class MultiOrState(SubsetState):
    """
    A state for many states to be combined together with an 'or' operation.

    This is meant to be used for cases where many subset states are meant to be
    combined together and provides significant performance enhancements compared
    to chaining individual OrStates
    """

    def __init__(self, states):
        super(MultiOrState, self).__init__()
        if len(states) < 1:
            raise ValueError("states should contain at least one subset state")
        self.states = states

    def copy(self):
        return type(self)(self.states)

    @property
    def attributes(self):
        att = self.states[0].attributes
        for state in self.states[1:]:
            att += state.attributes
        return tuple(sorted(set(att)))

    @memoize
    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):
        # Copy the first mask so that we can then modify it in-place
        result = self.states[0].to_mask(data, view=view).copy()
        for state in self.states[1:]:
            result |= state.to_mask(data, view=view)
        return result

    def __str__(self):
        return "('or' combination of {0} individual states)".format(len(self.states))


class MaskSubsetState(SubsetState):
    """
    A subset defined by a boolean mask.

    Parameters
    ----------
    mask : `~numpy.ndarray`
        The boolean mask to apply to the data.
    cids : iterable of :class:`~glue.core.component_id.ComponentID`
        The component IDs along which the mask applies.
    """

    def __init__(self, mask, cids):
        self._cids = cids
        self._mask = np.asarray(mask, dtype=bool)

    @property
    def mask(self):
        """
        The boolean mask to apply to the data.
        """
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value

    @property
    def cids(self):
        """
        The component IDs along which the mask applies.
        """
        return self._cids

    @cids.setter
    def cids(self, value):
        self._cids = value

    @property
    def attributes(self):
        return self._cids

    def copy(self):
        return MaskSubsetState(self.mask, self.cids)

    def to_mask(self, data, view=None):

        if view is None:
            view = slice(None)

        # shortcut for data on the same pixel grid
        if data.pixel_component_ids == self.cids:
            return self.mask[view].copy()

        # locate each element of data in the coordinate system of the mask
        vals = [data[c, view].astype(np.int) for c in self.cids]
        result = self.mask[tuple(vals)]

        for v, n in zip(vals, data.shape):
            result &= ((v >= 0) & (v < n))

        return result

    def __gluestate__(self, context):
        return dict(cids=[context.id(c) for c in self.cids],
                    mask=context.do(self.mask))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(context.object(rec['mask']),
                   [context.object(c) for c in rec['cids']])


class SliceSubsetState(SubsetState):
    """
    A subset defined by a set of array slices.

    Parameters
    ----------
    reference_data : :class:`~glue.core.data.Data`
        The data in whose space the slices are defined.
    slices : iterable of :class:`slice`
        An iterable containing :class:`slice` objects to apply to the data.
    """

    def __init__(self, reference_data, slices):
        self._reference_data = reference_data
        self._slices = slices
        self._pad_slices()

    @property
    def reference_data(self):
        """
        The data in whose space the slices are defined.
        """
        return self._reference_data

    @reference_data.setter
    def reference_data(self, value):
        self._reference_data = value

    @property
    def slices(self):
        """
        An iterable containing :class:`slice` objects to apply to the data.
        """
        return self._slices

    @slices.setter
    def slices(self, value):
        self._slices = value

    def _pad_slices(self):
        from glue.core.data import BaseCartesianData
        if isinstance(self.reference_data, BaseCartesianData) and len(self.slices) < self.reference_data.ndim:
            self.slices = self.slices + [slice(None)] * (self.reference_data.ndim - len(self.slices))

    @property
    def attributes(self):
        return self._reference_data.pixel_component_ids

    def copy(self):
        return SliceSubsetState(self.reference_data, self.slices)

    def to_mask(self, data, view=None):

        if view is None:
            view = Ellipsis
        elif isinstance(view, slice) or np.isscalar(view):
            view = (view,)

        # Figure out the shape of the final mask given the requested view
        shape = view_shape(data.shape, view)

        if data is self.reference_data:

            slices = self.slices

        else:

            # Check if we can transform list of slices to match this dataset
            order = data.pixel_aligned_data.get(self.reference_data, None)

            if order is None:
                # We use broadcast_to for minimal memory usage
                return broadcast_to(False, shape)
            else:
                # Reorder slices
                slices = [self.slices[idx] for idx in order]

        if (isinstance(view, np.ndarray) or
                (isinstance(view, (tuple, list)) and isinstance(view[0], np.ndarray))):
            mask = np.zeros(data.shape, dtype=bool)
            mask[tuple(slices)] = True
            return mask[view]

        # The original slices assume the full array, not the array with the view
        # applied, so we need to now adjust the slices accordingly.
        if view is Ellipsis:
            subslices = slices
        else:
            subslices = []
            for i in range(data.ndim):
                if i >= len(view):
                    subslices.append(slices[i])
                elif np.isscalar(view[i]):
                    beg, end, stp = slices[i].indices(data.shape[i])
                    if view[i] < beg or view[i] >= end or (view[i] - beg) % stp != 0:
                        return broadcast_to(False, shape)
                elif isinstance(view[i], slice):
                    if view[i].step is not None and view[i].step < 0:
                        beg, end, step = view[i].indices(data.shape[i])
                        v = slice(end + 1, beg + 1, -step)
                    else:
                        v = view[i]
                    subslices.append(combine_slices(v, slices[i], data.shape[i]))
                else:
                    raise TypeError("Unexpected view item: {0}".format(view[i]))

        # Create mask with final shape
        mask = np.zeros(shape, dtype=bool)
        mask[tuple(subslices)] = True

        return mask

    def to_array(self, data, att):
        if data is self.reference_data:
            slices = self.slices
        else:
            order = data.pixel_aligned_data.get(self.reference_data, None)
            if order is None:
                raise IncompatibleAttribute()
            slices = [self.slices[idx] for idx in order]
        return data[att, tuple(slices)]

    def __gluestate__(self, context):
        return dict(slices=context.do(self.slices),
                    reference_data=context.id(self.reference_data))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(rec['reference_data'], context.object(rec['slices']))

    def __setgluestate_callback__(self, context):
        self.reference_data = context.object(self.reference_data)
        self._pad_slices()


class CategorySubsetState(SubsetState):
    """
    A subset defined by the set of categorical values that are equal to
    a set of categories.

    Parameters
    ----------
    att : :class:`~glue.core.component_id.ComponentID`
        The categorical data attribute used for the subset
    categories : iterable
        The categories that the attribute should be equal to. These should be
        given as the integer codes, not the categorical labels.
    """

    def __init__(self, att, categories):
        super(CategorySubsetState, self).__init__()
        self._att = att
        self._categories = np.asarray(categories).ravel()

    @property
    def att(self):
        """
        The categorical data attribute used for the subset
        """
        return self._att

    @att.setter
    def att(self, value):
        self._att = value

    @property
    def categories(self):
        """
        The categories that the attribute should be equal to. These should be
        given as the integer codes, not the categorical labels.
        """
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

    @property
    def attributes(self):
        return self._att,

    @memoize
    def to_mask(self, data, view=None):
        vals = data[self._att, view]
        if isinstance(vals, categorical_ndarray):
            vals = vals.codes
        result = np.in1d(vals.ravel(), self._categories)
        return result.reshape(vals.shape)

    def copy(self):
        return CategorySubsetState(self._att, self._categories.copy())

    def __gluestate__(self, context):
        return dict(att=context.id(self._att),
                    vals=context.do(self._categories))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(context.object(rec['att']),
                   context.object(rec['vals']))


class ElementSubsetState(SubsetState):
    """
    A subset defined by a set of indices to apply to the data.

    Parameters
    ----------
    indices
        Any valid object that can be used to index a Numpy array.
    data : :class:`~glue.core.data.Data`
        The data in whose space the indices are defined.
    """

    def __init__(self, indices=None, data=None):
        super(ElementSubsetState, self).__init__()
        self._indices = indices
        if data is None:
            self._data_uuid = None
        else:
            self._data_uuid = data.uuid

    @property
    def indices(self):
        """
        The indices which when applied to the data give the subset.
        """
        return self._indices

    @indices.setter
    def indices(self, value):
        self._indices = value

    @property
    def data(self):
        """
        The UUID of the data in whose space the indices are defined.
        """
        return self._data_uuid

    @data.setter
    def data(self, value):
        self._data = value

    @memoize
    def to_mask(self, data, view=None):
        if data.uuid == self._data_uuid or self._data_uuid is None:
            # XXX this is inefficient for views
            result = np.zeros(data.shape, dtype=bool)
            if self._indices is not None:
                try:
                    result.flat[self._indices] = True
                except IndexError:
                    if self._data_uuid is None:
                        raise IncompatibleAttribute()
                    else:
                        raise
            if view is not None:
                result = result[view]
            return result
        else:
            raise IncompatibleAttribute()

    @property
    def attributes(self):
        return self._data.pixel_component_ids

    def copy(self):
        state = ElementSubsetState(indices=self._indices)
        state._data_uuid = self._data_uuid
        return state

    def __gluestate__(self, context):
        return dict(indices=context.do(self._indices),
                    data_uuid=self._data_uuid)

    @classmethod
    def __setgluestate__(cls, rec, context):
        state = cls(indices=context.object(rec['indices']))
        try:
            state._data_uuid = rec['data_uuid']
        except KeyError:  # BACKCOMPAT
            pass
        return state


VALID_INEQUALTIY_OPS = [operator.gt, operator.ge,
                        operator.lt, operator.le,
                        operator.eq, operator.ne]


class InequalitySubsetState(SubsetState):
    """
    A subset defined by a mathematical comparison of a attribute values to
    a reference value or attribute values.

    Parameters
    ----------
    left : float or `~glue.core.component_id.ComponentID` or str
        The value or component on the left hand side of the comparison.
    right : float or `~glue.core.component_id.ComponentID` or str
        The value or component on the right hand side of the comparison.
    operator : operator
        The comparison operator (from the :mod:`operator` module)
    """

    def __init__(self, left, right, operator):
        from glue.core.component_link import ComponentLink

        super(InequalitySubsetState, self).__init__()
        from glue.core.data import ComponentID
        if operator not in VALID_INEQUALTIY_OPS:
            raise TypeError("Invalid boolean operator: %s" % operator)
        if not isinstance(left, (ComponentID, numbers.Number,
                                 ComponentLink, six.string_types)):
            raise TypeError("Input must be ComponentID or NumberType or string: %s"
                            % type(left))

        if not isinstance(right, (ComponentID, numbers.Number,
                                  ComponentLink, six.string_types)):
            raise TypeError("Input must be ComponentID or NumberType or string: %s"
                            % type(right))
        self._left = left
        self._right = right
        self._operator = operator

    @property
    def left(self):
        """
        The value or component on the left hand side of the comparison.
        """
        return self._left

    @left.setter
    def left(self, value):
        self._left = value

    @property
    def right(self):
        """
        The value or component on the right hand side of the comparison.
        """
        return self._right

    @right.setter
    def right(self, value):
        self._right = value

    @property
    def operator(self):
        """
        The comparison operator (from the :mod:`operator` module)
        """
        return self._operator

    @operator.setter
    def operator(self, value):
        self._operator = value

    @memoize
    def to_mask(self, data, view=None):

        # FIXME: the default view in glue should be ... not None, because
        # if x is a Numpy array, x[None] has one more dimension than x. For
        # now we just fix this for the scope of this method.
        if view is None:
            view = Ellipsis

        if isinstance(self._left, (numbers.Number, six.string_types)):
            left = self._left
        else:
            left = data[self._left, view]

        if isinstance(self._right, (numbers.Number, six.string_types)):
            right = self._right
        else:
            right = data[self._right, view]

        return self._operator(left, right)

    def copy(self):
        return InequalitySubsetState(self._left, self._right, self._operator)

    def __str__(self):
        sym = OPSYM.get(self._operator, self._operator)
        return "(%s %s %s)" % (self._left, sym, self._right)

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self)


class FloodFillSubsetState(MaskSubsetState):
    """
    A subset representing a flood-fill operation, which is computed on-the-fly.

    Parameters
    ----------
    data : :class:`~glue.core.data.Data`
        The data on which the flood fill is computed.
    att : :class:`glue.core.component_id.ComponentID`
        The attribute defining the values to use for the flood fill.
    start_coords : tuple
        The pixel coordinates of the starting point.
    threshold : float
        A value greater or equal to 1 describing the extend of the flood
        filling. The range of values selected by the flood fill is ``start_value
        * (2 -threshold)`` to ``start_value * threshold`` where ``start_value``
        is the value of the data at ``start_coords``.
    """

    # TODO: we need to recompute the mask if the numerical values of the
    # data changes.

    def __init__(self, data, att, start_coords, threshold):

        if len(start_coords) != data.ndim:
            raise ValueError("start_coords should have as many values as data "
                             "has dimensions.")

        self._att = att
        self._data = data
        self._start_coords = tuple(start_coords)
        self._threshold = float(threshold)
        self._cids = self.data.pixel_component_ids

        self._compute_mask()

    @property
    def data(self):
        """
        The data on which the flood fill is computed.
        """
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def att(self):
        """
        The attribute defining the values to use for the flood fill.
        """
        return self._att

    @att.setter
    def att(self, value):
        self._att = value

    @property
    def start_coords(self):
        """
        The pixel coordinates of the starting point.
        """
        return self._start_coords

    @start_coords.setter
    def start_coords(self, value):
        self._start_coords = value

    @property
    def threshold(self):
        """
        A value greater or equal to 1 describing the extend of the flood
        filling. The range of values selected by the flood fill is ``start_value
        * (2 -threshold)`` to ``start_value * threshold`` where ``start_value``
        is the value of the data at ``start_coords``.
        """
        return self._threshold

    @threshold.setter
    def threshold(self, value):
        self._threshold = value

    def _compute_mask(self):
        mask = floodfill(self.data[self.att],
                         self.start_coords, self.threshold)
        self._mask_cache = (self._hash, mask)

    @property
    def _hash(self):
        return self.data, self.att, self.start_coords, self.threshold, self.cids

    @property
    def mask(self):
        if self._mask_cache[0] != self._hash:
            self._compute_mask()
        return self._mask_cache[1]

    @property
    def attributes(self):
        return list(self._data.pixel_component_ids) + [self.att]

    def copy(self):
        return FloodFillSubsetState(self.data, self.att, self.start_coords,
                                    self.threshold)

    def __gluestate__(self, context):
        # We don't store the data since this would cause a circular reference.
        # However we can recover the data from the attribute ComponentID.
        return dict(attribute=context.id(self.att),
                    start_coords=self.start_coords,
                    threshold=self.threshold)

    @classmethod
    def __setgluestate__(cls, rec, context):
        att = context.object(rec['attribute'])
        return cls(att.parent, att,
                   context.object(rec['start_coords']),
                   context.object(rec['threshold']))


class RoiSubsetState3d(SubsetState):
    """
    A subset defined as the set of points in three dimensions that lie inside
    a 3-d region of interest (ROI).

    The three dimensions are defined as three numerical data attributes.

    Parameters
    ----------
    xatt : :class:`~glue.core.component_id.ComponentID`
        The data attribute on the x axis.
    yatt : :class:`~glue.core.component_id.ComponentID`
        The data attribute on the y axis.
    zatt : :class:`~glue.core.component_id.ComponentID`
        The data attribute on the z axis.
    roi : :class:`~glue.core.roi.Roi`
        The region of interest (which should implement ``contains3d``)
    """

    @contract(xatt='isinstance(ComponentID)', yatt='isinstance(ComponentID)', zatt='isinstance(ComponentID)')
    def __init__(self, xatt=None, yatt=None, zatt=None, roi=None):
        super(RoiSubsetState3d, self).__init__()
        self._xatt = xatt
        self._yatt = yatt
        self._zatt = zatt
        self._roi = roi

    @property
    def xatt(self):
        """
        The data attribute on the x axis.
        """
        return self._xatt

    @xatt.setter
    def xatt(self, value):
        self._xatt = value

    @property
    def yatt(self):
        """
        The data attribute on the y axis.
        """
        return self._yatt

    @yatt.setter
    def yatt(self, value):
        self._yatt = value

    @property
    def zatt(self):
        """
        The data attribute on the z axis.
        """
        return self._zatt

    @zatt.setter
    def zatt(self, value):
        self._zatt = value

    @property
    def roi(self):
        """
        The region of interest.
        """
        return self._roi

    @roi.setter
    def roi(self, value):
        self._roi = value

    @property
    def attributes(self):
        return (self.xatt, self.yatt, self.zatt)

    @contract(data='isinstance(Data)', view='array_view')
    def to_mask(self, data, view=None):

        # TODO: make sure that pixel components don't actually take up much
        #       memory and are just views
        x = data[self.xatt, view]
        y = data[self.yatt, view]
        z = data[self.zatt, view]

        if self.roi.defined():
            result = self.roi.contains3d(x, y, z)
        else:
            result = np.zeros(x.shape, dtype=bool)

        if result.shape != x.shape:
            raise ValueError("Unexpected error: boolean mask has incorrect dimensions")

        return result

    def copy(self):
        result = RoiSubsetState3d()
        result.xatt = self.xatt
        result.yatt = self.yatt
        result.zatt = self.zatt
        result.roi = self.roi
        return result

    def __gluestate__(self, context):
        return dict(xatt=context.id(self.xatt),
                    yatt=context.id(self.yatt),
                    zatt=context.id(self.zatt),
                    roi=context.id(self.roi))

    @classmethod
    def __setgluestate__(cls, rec, context):
        return RoiSubsetState3d(context.object(rec['xatt']),
                                context.object(rec['yatt']),
                                context.object(rec['zatt']),
                                context.object(rec['roi']))


@contract(subsets='list(isinstance(Subset))', returns=Subset)
def _combine(subsets, operator):
    state = operator(*[s.subset_state for s in subsets])
    result = Subset(None)
    result.subset_state = state
    return result


def combine_multiple(subsets, operator):
    if len(subsets) == 0:
        return SubsetState()
    else:
        combined = subsets[0]
        for subset in subsets[1:]:
            combined = operator(combined, subset)
        return combined


def roi_to_subset_state(roi, x_att=None, y_att=None, x_categories=None, y_categories=None):
    """
    Given a 2D ROI and attributes on the x and y axis, determine the
    corresponding subset state.
    """

    if isinstance(roi, RangeROI):

        if roi.ori == 'x':
            att = x_att
            categories = x_categories
        else:
            att = y_att
            categories = y_categories

        if categories is not None:
            return CategoricalROISubsetState.from_range(categories, att, roi.min, roi.max)
        else:
            return RangeSubsetState(roi.min, roi.max, att)

    elif x_categories is not None or y_categories is not None:

        if isinstance(roi, RectangularROI):

            # In this specific case, we can decompose the rectangular ROI into
            # two RangeROIs that are combined with an 'and' logical operation.

            range1 = XRangeROI(roi.xmin, roi.xmax)
            range2 = YRangeROI(roi.ymin, roi.ymax)

            subset1 = roi_to_subset_state(range1, x_att=x_att, x_categories=x_categories)
            subset2 = roi_to_subset_state(range2, y_att=y_att, y_categories=y_categories)

            return AndState(subset1, subset2)

        elif isinstance(roi, CategoricalROI):

            # The selection is categorical itself. We assume this is along the x axis

            return CategoricalROISubsetState(roi=roi, att=x_att)

        else:

            # The selection is polygon-like, which requires special care.

            if x_categories is not None and y_categories is not None:

                # For each category, we check which categories along the other
                # axis fall inside the polygon:

                selection = {}

                for code, label in enumerate(x_categories):

                    # Determine the coordinates of the points to check
                    n_other = len(y_categories)
                    y = np.arange(n_other)
                    x = np.repeat(code, n_other)

                    # Determine which points are in the polygon, and which
                    # categories these correspond to
                    in_poly = roi.contains(x, y)
                    categories = y_categories[in_poly]

                    if len(categories) > 0:
                        selection[label] = set(categories)

                return CategoricalROISubsetState2D(selection, x_att, y_att)

            else:

                # If one of the components is not categorical, we treat this as
                # if each categorical component was mapped to a numerical value,
                # and at each value, we keep track of the polygon intersection
                # with the component. This will result in zero, one, or multiple
                # separate numerical ranges for each categorical value.

                # TODO: if we ever allow the category order to be changed, we
                # need to figure out how to update this!

                # We loop over each category and for each one we find the
                # numerical ranges

                selection = {}

                if x_categories is not None:
                    categories = x_categories
                    cat_att = x_att
                    num_att = y_att
                    x, y = roi.to_polygon()
                else:
                    categories = y_categories
                    cat_att = y_att
                    num_att = x_att
                    y, x = roi.to_polygon()

                for code, label in enumerate(categories):

                    # We determine all the numerical segments that represent the
                    # ensemble of points in y that fall in the polygon
                    # TODO: profile the following function
                    segments = polygon_line_intersections(x, y, xval=code)

                    if len(segments) > 0:
                        selection[label] = segments

                return CategoricalMultiRangeSubsetState(selection, cat_att=cat_att, num_att=num_att)

    else:

        # The selection is polygon-like and components are numerical

        subset_state = RoiSubsetState()
        subset_state.xatt = x_att
        subset_state.yatt = y_att
        subset_state.roi = PolygonalROI(*roi.to_polygon())

        return subset_state



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

from __future__ import absolute_import, division, print_function

from glue.viewers.matplotlib.qt.data_viewer import MatplotlibDataViewer
from glue.viewers.matplotlib.toolbar_mode import ToolbarModeBase
from glue.core.tests.util import simple_session


class ToolbarModeTest(ToolbarModeBase):

    tool_id = 'test'
    tool_tip = 'just testing'
    icon = 'glue_square'

    def __init__(self, axes, release_callback=None):
        super(ToolbarModeTest, self).__init__(axes, release_callback=release_callback)
        self.action_text = 'test text'
        self.last_mode = None

    def press(self, event):
        self.last_mode = 'PRESS'

    def move(self, event):
        self.last_mode = 'MOVE'


class ExampleViewer(MatplotlibDataViewer):

    def __init__(self, session, parent=None):
        super(ExampleViewer, self).__init__(session, parent=parent)
        self.axes.plot([1, 2, 3])[0]

    def initialize_toolbar(self):
        super(ExampleViewer, self).initialize_toolbar()
        self.tool = ToolbarModeTest(self, release_callback=self.callback)
        self.toolbar.add_tool(self.tool)

    def callback(self, mode):
        self._called_back = True


class TestToolbar(object):

    def setup_method(self, method):
        self.session = simple_session()
        self.viewer = ExampleViewer(self.session)
        self._called_back = False

    def teardown_method(self, method):
        self.viewer.close()

    def assert_valid_mode_state(self, target_mode):
        for tool_id in self.viewer.toolbar.actions:
            if tool_id == target_mode and self.viewer.toolbar.actions[tool_id].isCheckable():
                assert self.viewer.toolbar.actions[tool_id].isChecked()
                self.viewer.toolbar._active == target_mode
            else:
                assert not self.viewer.toolbar.actions[tool_id].isChecked()

    def test_callback(self):
        self.viewer.toolbar.actions['mpl:home'].trigger()
        self.viewer.tool.release(None)
        assert self.viewer._called_back

    def test_change_mode(self):

        self.viewer.toolbar.actions['mpl:pan'].toggle()
        assert self.viewer.toolbar.active_tool.tool_id == 'mpl:pan'
        assert self.viewer._mpl_nav.mode == 'pan/zoom'

        self.viewer.toolbar.actions['mpl:pan'].toggle()
        assert self.viewer.toolbar.active_tool is None
        assert self.viewer._mpl_nav.mode == ''

        self.viewer.toolbar.actions['mpl:zoom'].trigger()
        assert self.viewer.toolbar.active_tool.tool_id == 'mpl:zoom'
        assert self.viewer._mpl_nav.mode == 'zoom rect'

        self.viewer.toolbar.actions['test'].trigger()
        assert self.viewer.toolbar.active_tool.tool_id == 'test'
        assert self.viewer._mpl_nav.mode == ''



from __future__ import absolute_import, division, print_function

from logging import getLogger, basicConfig, NullHandler
basicConfig()
logger = getLogger("glue")
# Default to Null unless we override this later
logger.addHandler(NullHandler())



from __future__ import absolute_import, division, print_function

from collections import OrderedDict
import pytest

from numpy.testing import assert_equal

from astropy.io import fits
from glue.io.formats.fits.subset_mask import fits_subset_mask_importer, fits_subset_mask_exporter


def test_reader(tmpdir):

    mask_filename = tmpdir.join('subset_mask.fits').strpath

    original_mask = [[0, 1], [1, 0]]
    hdu = fits.PrimaryHDU(data=original_mask)
    hdu.writeto(mask_filename)

    masks = fits_subset_mask_importer(mask_filename)

    assert len(masks) == 1
    label, mask = list(masks.items())[0]

    assert label == 'subset_mask'
    assert_equal(mask, original_mask)


def test_reader_extensions(tmpdir):

    mask_filename = tmpdir.join('subset_mask.fits').strpath

    original_mask1 = [[0, 1], [1, 0]]
    hdu1 = fits.ImageHDU(data=original_mask1, name='Subset A')
    original_mask2 = [[1, 0], [1, 0]]
    hdu2 = fits.ImageHDU(data=original_mask2)

    hdulist = fits.HDUList([fits.PrimaryHDU(), hdu1, hdu2])
    hdulist.writeto(mask_filename)

    masks = fits_subset_mask_importer(mask_filename)

    assert len(masks) == 2
    mask_items = list(masks.items())
    label1, mask1 = mask_items[0]
    label2, mask2 = mask_items[1]

    assert label1 == 'SUBSET A'
    assert_equal(mask1, original_mask1)

    assert label2 == 'subset_mask[2]'
    assert_equal(mask2, original_mask2)


def test_reader_invalid_hdus(tmpdir):

    mask_filename = tmpdir.join('subset_mask.fits').strpath

    hdu = fits.PrimaryHDU(data=[1.2, 3.2, 4.5])
    hdu.writeto(mask_filename)

    with pytest.raises(ValueError) as exc:
        fits_subset_mask_importer(mask_filename)
    assert exc.value.args[0] == 'No HDUs with integer values (which would normally indicate a mask) were found in file'


def test_reader_invalid_format(tmpdir):

    mask_filename = tmpdir.join('subset_mask.fits').strpath

    with open(mask_filename, 'w') as f:
        f.write('qwdiwqoidqwjdijwq')

    with pytest.raises(IOError) as exc:
        fits_subset_mask_importer(mask_filename)
    assert exc.value.args[0].endswith('is not a valid FITS file')


def test_writer(tmpdir):

    mask_filename = tmpdir.join('subset_mask.fits').strpath

    masks = OrderedDict()
    masks['subset 1'] = [[0, 1], [1, 0]]
    masks['subset 2'] = [[0, 1, 3], [1, 0, 2]]

    fits_subset_mask_exporter(mask_filename, masks)

    with fits.open(mask_filename) as hdulist:

        assert len(hdulist) == 3

        assert hdulist[0].data is None

        assert hdulist[1].name == 'SUBSET 1'
        assert_equal(hdulist[1].data, masks['subset 1'])

        assert hdulist[2].name == 'SUBSET 2'
        assert_equal(hdulist[2].data, masks['subset 2'])



from __future__ import absolute_import, division, print_function

import numpy as np
from mock import MagicMock
from numpy.testing import assert_allclose

from glue.core import Data
from glue.viewers.image.qt import StandaloneImageViewer, ImageViewer
from glue.tests.helpers import requires_astropy, requires_scipy
from glue.app.qt import GlueApplication
from glue.utils.qt import process_events

from ..pv_slicer import _slice_from_path, _slice_label, _slice_index, PVSliceWidget


@requires_astropy
@requires_scipy
class TestSliceExtraction(object):

    def setup_method(self, method):
        self.x = np.random.random((2, 3, 4))
        self.d = Data(x=self.x)

    def test_constant_y(self):

        slc = (0, 'y', 'x')
        x = [-0.5, 3.5]
        y = [0, 0]
        s = _slice_from_path(x, y, self.d, 'x', slc)[0]
        assert_allclose(s, self.x[:, 0, :])

    def test_constant_x(self):

        slc = (0, 'y', 'x')
        y = [-0.5, 2.5]
        x = [0, 0]
        s = _slice_from_path(x, y, self.d, 'x', slc)[0]
        assert_allclose(s, self.x[:, :, 0])

    def test_transpose(self):
        slc = (0, 'x', 'y')
        y = [-0.5, 3.5]
        x = [0, 0]
        s = _slice_from_path(x, y, self.d, 'x', slc)[0]
        assert_allclose(s, self.x[:, 0, :])


def test_slice_label():
    d = Data(x=np.zeros((2, 3, 4)))
    assert _slice_label(d, (0, 'y', 'x')) == 'World 0'
    assert _slice_label(d, ('y', 0, 'x')) == 'World 1'
    assert _slice_label(d, ('y', 'x', 0)) == 'World 2'


def test_slice_index():
    d = Data(x=np.zeros((2, 3, 4)))
    assert _slice_index(d, (0, 'y', 'x')) == 0
    assert _slice_index(d, ('y', 0, 'x')) == 1


class TestStandaloneImageViewer(object):

    def setup_method(self, method):
        im = np.random.random((3, 3))
        self.w = StandaloneImageViewer(im)

    def teardown_method(self, method):
        self.w.close()

    def test_set_cmap(self):
        cm_mode = self.w.toolbar.tools['image:colormap']
        act = cm_mode.menu_actions()[1]
        act.trigger()
        assert self.w._composite.layers['image']['color'] is act.cmap

    def test_double_set_image(self):
        assert len(self.w._axes.images) == 1
        self.w.set_image(np.zeros((3, 3)))
        assert len(self.w._axes.images) == 1


class MockImageViewer(object):

    def __init__(self, slice, data):
        self.slice = slice
        self.data = data
        self.wcs = None
        self.state = MagicMock()


class TestPVSliceWidget(object):

    def setup_method(self, method):

        self.d = Data(x=np.zeros((2, 3, 4)))
        self.slc = (0, 'y', 'x')
        self.image = MockImageViewer(self.slc, self.d)
        self.w = PVSliceWidget(image=np.zeros((3, 4)), wcs=None, image_viewer=self.image)

    def teardown_method(self, method):
        self.w.close()

    def test_basic(self):
        pass


class TestPVSliceTool(object):

    def setup_method(self, method):
        self.cube = Data(label='cube', x=np.arange(1000).reshape((5, 10, 20)))
        self.application = GlueApplication()
        self.application.data_collection.append(self.cube)
        self.viewer = self.application.new_data_viewer(ImageViewer)
        self.viewer.add_data(self.cube)

    def teardown_method(self, method):
        self.viewer.close()
        self.viewer = None
        self.application.close()
        self.application = None

    @requires_astropy
    @requires_scipy
    def test_basic(self):

        self.viewer.toolbar.active_tool = 'slice'

        self.viewer.axes.figure.canvas.draw()
        process_events()

        x, y = self.viewer.axes.transData.transform([[0.9, 4]])[0]
        self.viewer.axes.figure.canvas.button_press_event(x, y, 1)
        x, y = self.viewer.axes.transData.transform([[7.2, 6.6]])[0]
        self.viewer.axes.figure.canvas.button_press_event(x, y, 1)

        process_events()

        assert len(self.application.tab().subWindowList()) == 1

        self.viewer.axes.figure.canvas.key_press_event('enter')

        process_events()

        assert len(self.application.tab().subWindowList()) == 2

        pv_widget = self.application.tab().subWindowList()[1].widget()
        assert pv_widget._x.shape == (6,)
        assert pv_widget._y.shape == (6,)



from __future__ import absolute_import, division, print_function

import os

from qtpy import QtWidgets

from glue.config import link_function, link_helper
from glue.utils.decorators import avoid_circular
from glue.utils.qt import load_ui
from glue.external.echo.qt import autoconnect_callbacks_to_qt
from glue.external.echo.qt.connect import UserDataWrapper, connect_combo_selection
from glue.dialogs.link_editor.state import LinkEditorState

__all__ = ['LinkEditor', 'main']

N_COMBO_MAX = 10


def get_function_name(info):
    item = info[0]
    if hasattr(item, 'display') and item.display is not None:
        return item.display
    else:
        return item.__name__


class LinkMenu(QtWidgets.QMenu):

    def __init__(self, parent=None):

        super(LinkMenu, self).__init__(parent=parent)

        categories = []
        for function in link_function.members:
            if len(function.output_labels) == 1:
                categories.append(function.category)
        for helper in link_helper.members:
            categories.append(helper.category)
        categories = ['General'] + sorted(set(categories) - set(['General']))

        for category in categories:
            submenu = self.addMenu(category)
            for function in link_function.members:
                if function.category == category and len(function.output_labels) == 1:
                    action = submenu.addAction(get_function_name(function))
                    action.setData(UserDataWrapper(function))
            for helper in link_helper.members:
                if helper.category == category:
                    action = submenu.addAction(get_function_name(helper))
                    action.setData(UserDataWrapper(helper))


class LinkEditorWidget(QtWidgets.QWidget):

    def __init__(self, data_collection, suggested_links=None, parent=None):

        super(LinkEditorWidget, self).__init__(parent=parent)

        self._data_collection = data_collection

        self.state = LinkEditorState(data_collection, suggested_links=suggested_links)

        self._connections = []

        self._ui = load_ui('link_editor_widget.ui', self,
                           directory=os.path.dirname(__file__))
        self._handlers = autoconnect_callbacks_to_qt(self.state, self._ui)

        self._set_up_combos()

        self._ui.graph_widget.set_data_collection(data_collection, new_links=self.state.links)
        self._ui.graph_widget.selection_changed.connect(self._on_data_change_graph)

        self._menu = LinkMenu(parent=self._ui.button_add_link)
        self._menu.triggered.connect(self._add_link)
        self._ui.button_add_link.setMenu(self._menu)

        self._watched_links = []

        self.state.add_callback('data1', self._on_data_change)
        self.state.add_callback('data2', self._on_data_change)
        self._on_data_change()

        self.state.add_callback('data1', self._on_data_change_always)
        self.state.add_callback('data2', self._on_data_change_always)
        self._on_data_change_always()

        self.state.add_callback('current_link', self._on_current_link_change)
        self._on_current_link_change()

    def _add_link(self, action):
        self.state.new_link(action.data().data)

    def _set_up_combos(self):

        # Set up combo boxes - for now we hard-code the maximum number, but
        # we could do this more smartly by checking existing links and all
        # possible links in registry to figure out max number needed.

        self.att_names1 = []
        self.att_combos1 = []

        for combo_idx in range(N_COMBO_MAX):
            label_widget = QtWidgets.QLabel()
            combo_widget = QtWidgets.QComboBox(parent=self._ui)
            self.att_names1.append(label_widget)
            self.att_combos1.append(combo_widget)
            self._ui.combos1.addWidget(label_widget, combo_idx, 0)
            self._ui.combos1.addWidget(combo_widget, combo_idx, 1)

        self.att_names2 = []
        self.att_combos2 = []

        for combo_idx in range(N_COMBO_MAX):
            label_widget = QtWidgets.QLabel()
            combo_widget = QtWidgets.QComboBox(parent=self._ui)
            self.att_names2.append(label_widget)
            self.att_combos2.append(combo_widget)
            self._ui.combos2.addWidget(label_widget, combo_idx, 0)
            self._ui.combos2.addWidget(combo_widget, combo_idx, 1)

    @avoid_circular
    def _on_attribute_combo_change(self, *args, **kwargs):
        # Force a re-sync of the choices
        self._handlers['listsel_current_link'].update_widget(self.state.current_link, force=True)

    @avoid_circular
    def _on_data_change_graph(self):
        self.state.data1 = getattr(self._ui.graph_widget.selected_node1, 'data', None)
        self.state.data2 = getattr(self._ui.graph_widget.selected_node2, 'data', None)

    @avoid_circular
    def _on_data_change(self, *args):
        self._ui.graph_widget.manual_select(self.state.data1, self.state.data2)

    def _on_data_change_always(self, *args):
        # This should always run even when the change comes from the graph
        enabled = self.state.data1 is not None and self.state.data2 is not None
        self._ui.button_add_link.setEnabled(enabled)
        self._ui.button_simple_link.setEnabled(enabled)

    def _on_current_link_change(self, *args):

        # We update the link details panel on the right

        for connnection in self._connections:
            connnection.disconnect()

        self._connections = []

        link = self.state.current_link

        if link is None:
            self._ui.button_remove_link.setEnabled(False)
            self._ui.link_details.setText('')
            self._ui.combos1_header.hide()
            self._ui.combos2_header.hide()
            for widget in self.att_combos1 + self.att_names1 + self.att_combos2 + self.att_names2:
                widget.hide()
            return

        self._ui.button_remove_link.setEnabled(True)

        self._ui.link_details.setText(link.description)

        if link.data1 is self.state.data1:
            data1_names = link.names1
        else:
            data1_names = link.names2

        for idx, (label, combo) in enumerate(zip(self.att_names1, self.att_combos1)):
            if idx < len(data1_names):
                combo.show()
                label.show()
                label.setText(data1_names[idx])
                disconnector = connect_combo_selection(link, data1_names[idx], combo)
                self._connections.append(disconnector)
            else:
                label.hide()
                combo.hide()

        if link.data1 is self.state.data2:
            data2_names = link.names1
        else:
            data2_names = link.names2

        for idx, (label, combo) in enumerate(zip(self.att_names2, self.att_combos2)):
            if idx < len(data2_names):
                combo.show()
                label.show()
                label.setText(data2_names[idx])
                disconnector = connect_combo_selection(link, data2_names[idx], combo)
                self._connections.append(disconnector)
            else:
                label.hide()
                combo.hide()

        # Headers aren't needed if data2_names is 0 (legacy mode for old link
        # helpers where all attributes are 'inputs')
        if len(data2_names) == 0:
            self._ui.combos1_header.hide()
            self._ui.combos2_header.hide()
        else:
            self._ui.combos1_header.show()
            self._ui.combos2_header.show()

        self._ui.graph_widget.set_links(self.state.links)

        # When the user changes one of the attributes, we need to update the
        # main list of links to show the attributes being linked. This is
        # actually tricker than it sounds, and to solve this we listen for
        # changes in any of the link properties. We don't need to unsubscribe
        # since there's no harm in keeping them connected, and we only need to
        # subscribe to links that have been shown at least once as the other
        # ones can't be changed by users.
        if link not in self._watched_links:
            link.add_global_callback(self._on_attribute_combo_change)
            self._watched_links.append(link)


class LinkEditor(QtWidgets.QDialog):

    def __init__(self, data_collection, suggested_links=None, parent=None):

        super(LinkEditor, self).__init__(parent=parent)

        self._ui = load_ui('link_editor_dialog.ui', self,
                           directory=os.path.dirname(__file__))

        self.link_widget = LinkEditorWidget(data_collection,
                                              suggested_links=suggested_links,
                                              parent=self)

        self._ui.layout().insertWidget(1, self.link_widget)

    def accept(self, *args):
        self.link_widget.state.update_links_in_collection()
        super(LinkEditor, self).accept(*args)

    @classmethod
    def update_links(cls, collection, suggested_links=None):
        widget = cls(collection, suggested_links=suggested_links)
        widget._ui.exec_()


def main():  # pragma: no cover
    import numpy as np
    from glue.main import load_plugins
    from glue.utils.qt import get_qapp
    from glue.core import Data, DataCollection

    load_plugins()

    app = get_qapp()

    dc = DataCollection()

    for i in range(10):
        x = np.array([1, 2, 3])
        d = Data(label='data_{0:02d}'.format(i), x=x, y=x * 2)
        dc.append(d)

    LinkEditor.update_links(dc)



# The functions in this module are used to connect callback properties to Qt
# widgets.

from __future__ import absolute_import, division, print_function

import math
from functools import partial

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import Qt

import numpy as np

from ..core import add_callback, remove_callback
from ..selection import SelectionCallbackProperty, ChoiceSeparator

__all__ = ['connect_checkable_button', 'connect_text', 'connect_combo_data',
           'connect_combo_text', 'connect_float_text', 'connect_value',
           'connect_combo_selection', 'connect_list_selection',
           'BaseConnection']


class UserDataWrapper(object):
    def __init__(self, data):
        self.data = data


class BaseConnection(object):

    def __init__(self, instance, prop, widget):

        self._instance = instance
        self._prop = prop
        self._widget = widget


class connect_checkable_button(BaseConnection):
    """
    Connect a boolean callback property with a Qt button widget.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QtWidget
        The Qt widget to connect. This should implement the ``setChecked``
        method and the ``toggled`` signal.
    """

    def __init__(self, instance, prop, widget):
        super(connect_checkable_button, self).__init__(instance, prop, widget)
        self.connect()

    def update_widget(self, value):
        self._widget.setChecked(value)

    def update_prop(self, value):
        setattr(self._instance, self._prop, value)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.toggled.connect(self.update_prop)
        self._widget.setChecked(getattr(self._instance, self._prop) or False)

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.toggled.disconnect(self.update_prop)


class connect_text(BaseConnection):
    """
    Connect a string callback property with a Qt widget containing text.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QtWidget
        The Qt widget to connect. This should implement the ``setText`` and
        ``text`` methods as well optionally the ``editingFinished`` signal.
    """

    def __init__(self, instance, prop, widget):
        super(connect_text, self).__init__(instance, prop, widget)
        self.connect()

    def update_prop(self):
        value = self._widget.text()
        setattr(self._instance, self._prop, value)

    def update_widget(self, value):
        if hasattr(self._widget, 'editingFinished'):
            self._widget.blockSignals(True)
            self._widget.setText(value)
            self._widget.blockSignals(False)
            self._widget.editingFinished.emit()
        else:
            self._widget.setText(value)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        try:
            self._widget.editingFinished.connect(self.update_prop)
        except AttributeError:
            pass
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        try:
            self._widget.editingFinished.disconnect(self.update_prop)
        except AttributeError:
            pass


class connect_combo_data(BaseConnection):
    """
    Connect a callback property with a QComboBox widget based on the userData.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QComboBox
        The combo box to connect.

    See Also
    --------
    connect_combo_text: connect a callback property with a QComboBox widget based on the text.
    """

    def __init__(self, instance, prop, widget):
        super(connect_combo_data, self).__init__(instance, prop, widget)
        self.connect()

    def update_widget(self, value):
        try:
            idx = _find_combo_data(self._widget, value)
        except ValueError:
            if value is None:
                idx = -1
            else:
                raise
        self._widget.setCurrentIndex(idx)

    def update_prop(self, idx):
        if idx == -1:
            setattr(self._instance, self._prop, None)
        else:
            data_wrapper = self._widget.itemData(idx)
            if data_wrapper is None:
                setattr(self._instance, self._prop, None)
            else:
                setattr(self._instance, self._prop, data_wrapper.data)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.connect(self.update_prop)
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.disconnect(self.update_prop)


class connect_combo_text(BaseConnection):
    """
    Connect a callback property with a QComboBox widget based on the text.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QComboBox
        The combo box to connect.

    See Also
    --------
    connect_combo_data: connect a callback property with a QComboBox widget based on the userData.
    """

    def __init__(self, instance, prop, widget):
        super(connect_combo_text, self).__init__(instance, prop, widget)
        self.connect()

    def update_widget(self, value):
        try:
            idx = _find_combo_text(self._widget, value)
        except ValueError:
            if value is None:
                idx = -1
            else:
                raise
        self._widget.setCurrentIndex(idx)

    def update_prop(self, idx):
        if idx == -1:
            setattr(self._instance, self._prop, None)
        else:
            setattr(self._instance, self._prop, self._widget.itemText(idx))

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.connect(self.update_prop)
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.disconnect(self.update_prop)


class connect_float_text(BaseConnection):
    """
    Connect a numerical callback property with a Qt widget containing text.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QtWidget
        The Qt widget to connect. This should implement the ``setText`` and
        ``text`` methods as well optionally the ``editingFinished`` signal.
    fmt : str or func
        This should be either a format string (in the ``{}`` notation), or a
        function that takes a number and returns a string.
    """

    def __init__(self, instance, prop, widget, fmt="{:g}"):

        super(connect_float_text, self).__init__(instance, prop, widget)

        if callable(fmt):
            format_func = fmt
        else:
            def format_func(x):
                try:
                    return fmt.format(x)
                except ValueError:
                    return str(x)

        self._format_func = format_func

        self.connect()

    def update_prop(self):
        value = self._widget.text()
        try:
            value = float(value)
        except ValueError:
            try:
                value = np.datetime64(value)
            except Exception:
                value = 0
        setattr(self._instance, self._prop, value)

    def update_widget(self, value):
        if value is None:
            value = 0.
        self._widget.setText(self._format_func(value))

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        try:
            self._widget.editingFinished.connect(self.update_prop)
        except AttributeError:
            pass
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        try:
            self._widget.editingFinished.disconnect(self.update_prop)
        except AttributeError:
            pass


class connect_value(BaseConnection):
    """
    Connect a numerical callback property with a Qt widget representing a value.

    Parameters
    ----------
    instance : object
        The class instance that the callback property is attached to
    prop : str
        The name of the callback property
    widget : QtWidget
        The Qt widget to connect. This should implement the ``setText`` and
        ``text`` methods as well optionally the ``editingFinished`` signal.
    value_range : iterable, optional
        A pair of two values representing the true range of values (since
        Qt widgets such as sliders can only have values in certain ranges).
    log : bool, optional
        Whether the Qt widget value should be mapped to the log of the callback
        property.
    """

    def __init__(self, instance, prop, widget, value_range=None, log=False):

        super(connect_value, self).__init__(instance, prop, widget)

        if log:
            if value_range is None:
                raise ValueError("log option can only be set if value_range is given")
            else:
                self._value_range = math.log10(value_range[0]), math.log10(value_range[1])
        else:
            self._value_range = value_range
        self._log = log

        self.connect()

    def update_prop(self):
        value = self._widget.value()
        if self._value_range is not None:
            imin, imax = self._widget.minimum(), self._widget.maximum()
            value = (value - imin) / (imax - imin) * (self._value_range[1] - self._value_range[0]) + self._value_range[0]
        if self._log:
            value = 10 ** value
        setattr(self._instance, self._prop, value)

    def update_widget(self, value):
        if value is None:
            self._widget.setValue(0)
            return
        if self._log:
            value = math.log10(value)
        if self._value_range is not None:
            imin, imax = self._widget.minimum(), self._widget.maximum()
            value = (value - self._value_range[0]) / (self._value_range[1] - self._value_range[0]) * (imax - imin) + imin
        self._widget.setValue(value)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.valueChanged.connect(self.update_prop)
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.valueChanged.disconnect(self.update_prop)


class connect_button(BaseConnection):
    """
    Connect a button with a callback method

    Parameters
    ----------
    instance : object
        The class instance that the callback method is attached to
    prop : str
        The name of the callback method
    widget : QtWidget
        The Qt widget to connect. This should implement the ``clicked`` method
    """

    def __init__(self, instance, prop, widget):
        super(connect_button, self).__init__(instance, prop, widget)
        self.connect()

    def connect(self):
        self._widget.clicked.connect(getattr(self._instance, self._prop))

    def disconnect(self):
        self._widget.clicked.disconnect(self.update_prop)


def _find_combo_data(widget, value):
    """
    Returns the index in a combo box where itemData == value

    Raises a ValueError if data is not found
    """
    # Here we check that the result is True, because some classes may overload
    # == and return other kinds of objects whether true or false.
    for idx in range(widget.count()):
        if widget.itemData(idx) is not None:
            if isinstance(widget.itemData(idx), UserDataWrapper):
                if widget.itemData(idx).data is value or (widget.itemData(idx).data == value) is True:
                    return idx
            else:
                if widget.itemData(idx) is value or (widget.itemData(idx) == value) is True:
                    return idx
    else:
        raise ValueError("%s not found in combo box" % (value,))


def _find_combo_text(widget, value):
    """
    Returns the index in a combo box where text == value

    Raises a ValueError if data is not found
    """
    i = widget.findText(value)
    if i == -1:
        raise ValueError("%s not found in combo box" % value)
    else:
        return i


class connect_combo_selection(BaseConnection):

    def __init__(self, instance, prop, widget):

        if not isinstance(getattr(type(instance), prop), SelectionCallbackProperty):
            raise TypeError('connect_combo_selection requires a SelectionCallbackProperty')

        super(connect_combo_selection, self).__init__(instance, prop, widget)
        self.connect()

    def update_widget(self, value):

        # Update choices in the combo box

        combo_data = [self._widget.itemData(idx) for idx in range(self._widget.count())]
        combo_text = [self._widget.itemText(idx) for idx in range(self._widget.count())]

        choices = getattr(type(self._instance), self._prop).get_choices(self._instance)
        choice_labels = getattr(type(self._instance), self._prop).get_choice_labels(self._instance)

        if combo_data == choices and combo_text == choice_labels:
            choices_updated = False
        else:

            self._widget.blockSignals(True)
            self._widget.clear()

            if len(choices) == 0:
                return

            combo_model = self._widget.model()

            for index, (label, choice) in enumerate(zip(choice_labels, choices)):

                self._widget.addItem(label, userData=UserDataWrapper(choice))

                # We interpret None data as being disabled rows (used for headers)
                if isinstance(choice, ChoiceSeparator):
                    item = combo_model.item(index)
                    palette = self._widget.palette()
                    item.setFlags(item.flags() & ~(Qt.ItemIsSelectable | Qt.ItemIsEnabled))
                    item.setData(palette.color(QtGui.QPalette.Disabled, QtGui.QPalette.Text))

            choices_updated = True

        # Update current selection
        try:
            idx = _find_combo_data(self._widget, value)
        except ValueError:
            if value is None:
                idx = -1
            else:
                raise

        if idx == self._widget.currentIndex() and not choices_updated:
            return

        self._widget.setCurrentIndex(idx)
        self._widget.blockSignals(False)
        self._widget.currentIndexChanged.emit(idx)

    def update_prop(self, idx):
        if idx == -1:
            setattr(self._instance, self._prop, None)
        else:
            data_wrapper = self._widget.itemData(idx)
            if data_wrapper is None:
                setattr(self._instance, self._prop, None)
            else:
                setattr(self._instance, self._prop, data_wrapper.data)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.connect(self.update_prop)
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.currentIndexChanged.disconnect(self.update_prop)


class connect_list_selection(BaseConnection):

    def __init__(self, instance, prop, widget):
        """
        Connect a SelectionCallbackProperty with a QListWidget that supports
        single-item selection.
        """

        if not isinstance(getattr(type(instance), prop), SelectionCallbackProperty):
            raise TypeError('connect_list_selection requires a SelectionCallbackProperty')

        super(connect_list_selection, self).__init__(instance, prop, widget)
        self.connect()

    def update_widget(self, value, force=False):

        items = [self._widget.item(idx) for idx in range(self._widget.count())]
        list_text = [item.text() for item in items]
        list_data = [item.data(Qt.UserRole) for item in items]
        list_data = [d.data if d is not None else d for d in list_data]

        choices = getattr(type(self._instance), self._prop).get_choices(self._instance)
        choice_labels = getattr(type(self._instance), self._prop).get_choice_labels(self._instance)

        for idx in range(len(choices)):
            if choices[idx] is value:
                break
        else:
            idx = -1

        self._widget.blockSignals(True)

        choices_match = list_data == choices and list_text == choice_labels

        if force or not choices_match:

            self._widget.clear()

            if len(choices) == 0:
                self._widget.blockSignals(False)
                return

            for index, (label, choice) in enumerate(zip(choice_labels, choices)):

                item = QtWidgets.QListWidgetItem(label)
                item.setData(Qt.UserRole, UserDataWrapper(choice))
                self._widget.addItem(item)

                # We interpret None data as being disabled rows (used for headers)
                if isinstance(choice, ChoiceSeparator):
                    palette = self._widget.palette()
                    item.setFlags(item.flags() & ~(Qt.ItemIsSelectable | Qt.ItemIsEnabled))
                    # item.setData(palette.color(QtGui.QPalette.Disabled, QtGui.QPalette.Text))

        if len(self._widget.selectedItems()) == 0:
            current_index = -1
        else:
            selected_item = self._widget.selectedItems()[0]
            for current_index, item in enumerate(items):
                if item is selected_item:
                    break
            else:
                current_index = -1

        if idx == current_index and choices_match:
            self._widget.blockSignals(False)
            return

        self._widget.setCurrentItem(self._widget.item(idx))
        self._widget.blockSignals(False)
        self._widget.itemSelectionChanged.emit()

    def update_prop(self):

        if len(self._widget.selectedItems()) == 0:
            setattr(self._instance, self._prop, None)
        else:
            data_wrapper = self._widget.selectedItems()[0].data(Qt.UserRole)
            if data_wrapper is None:
                setattr(self._instance, self._prop, None)
            else:
                setattr(self._instance, self._prop, data_wrapper.data)

    def connect(self):
        add_callback(self._instance, self._prop, self.update_widget)
        self._widget.itemSelectionChanged.connect(self.update_prop)
        self.update_widget(getattr(self._instance, self._prop))

    def disconnect(self):
        remove_callback(self._instance, self._prop, self.update_widget)
        self._widget.itemSelectionChanged.disconnect(self.update_prop)



"""
Matplotlib artist for fixed resolution buffers.
"""

from __future__ import print_function, division

from mpl_scatter_density.base_image_artist import BaseImageArtist

import numpy as np

from glue.utils import color2rgb


class FRBArtist(BaseImageArtist):

    def __init__(self, ax, **kwargs):
        self._array_maker = None
        super(FRBArtist, self).__init__(ax, update_while_panning=False,
                                        array_func=self.array_func_wrapper,
                                        **kwargs)

    def array_func_wrapper(self, bins=None, range=None):
        if self._array_maker is None:
            return np.array([[np.nan]])
        else:
            ny, nx = bins
            (ymin, ymax), (xmin, xmax) = range
            bounds = [(ymin, ymax, ny), (xmin, xmax, nx)]
            array = self._array_maker(bounds)
            if array is None:
                return np.array([[np.nan]])
            else:
                return array

    def set_array_maker(self, array_maker):
        self._array_maker = array_maker

    def invalidate_cache(self):
        self.stale = True
        self.set_visible(True)


def imshow(axes, array_maker, aspect=None, vmin=None, vmax=None, color=None, **kwargs):
    """
    Similar to matplotlib's imshow command, but produces a FRBArtist
    """

    axes.set_aspect(aspect)

    im = FRBArtist(axes, **kwargs)

    if color:

        def wrapper(bounds=None):

            # Get original array
            mask = array_maker(bounds=bounds)

            # Convert to RGBA array"
            r, g, b = color2rgb(color)
            mask = np.dstack((r * mask, g * mask, b * mask, mask * .5))
            mask = (255 * mask).astype(np.uint8)

            return mask

        im.set_array_maker(wrapper)

    else:

        im.set_array_maker(array_maker)

    axes._set_artist_props(im)

    if im.get_clip_path() is None:
        # image does not already have clipping set, clip to axes patch
        im.set_clip_path(axes.patch)

    if vmin is not None or vmax is not None:
        im.set_clim(vmin, vmax)

    axes.images.append(im)

    def remove(h):
        axes.images.remove(h)

    # The following is needed for Matplotlib's .remove() method to work
    im._remove_method = remove

    return im



from __future__ import absolute_import, division, print_function

import os

import numpy as np

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt

from glue.external.echo.qt import autoconnect_callbacks_to_qt, connect_value
from glue.utils.qt import load_ui, fix_tab_widget_fontsize


class ScatterLayerStyleEditor(QtWidgets.QWidget):

    def __init__(self, layer, parent=None):

        super(ScatterLayerStyleEditor, self).__init__(parent=parent)

        self.ui = load_ui('layer_style_editor.ui', self,
                          directory=os.path.dirname(__file__))

        connect_kwargs = {'alpha': dict(value_range=(0, 1)),
                          'size_scaling': dict(value_range=(0.1, 10), log=True),
                          'density_contrast': dict(value_range=(0, 1)),
                          'vector_scaling': dict(value_range=(0.1, 10), log=True)}
        self._connections = autoconnect_callbacks_to_qt(layer.state, self.ui, connect_kwargs)

        self._connection_dpi = connect_value(layer.state.viewer_state, 'dpi', self.ui.value_dpi,
                                             value_range=(12, 144), log=True)

        fix_tab_widget_fontsize(self.ui.tab_widget)

        self.layer_state = layer.state

        self.layer_state.add_callback('markers_visible', self._update_markers_visible)
        self.layer_state.add_callback('line_visible', self._update_line_visible)
        self.layer_state.add_callback('xerr_visible', self._update_xerr_visible)
        self.layer_state.add_callback('yerr_visible', self._update_yerr_visible)
        self.layer_state.add_callback('vector_visible', self._update_vectors_visible)

        self.layer_state.add_callback('cmap_mode', self._update_cmap_mode)
        self.layer_state.add_callback('size_mode', self._update_size_mode)
        self.layer_state.add_callback('vector_mode', self._update_vector_mode)

        self.layer_state.add_callback('density_map', self._update_size_mode)
        self.layer_state.add_callback('density_map', self._update_warnings)
        self.layer_state.add_callback('density_map', self._update_checkboxes)

        self.layer_state.add_callback('layer', self._update_warnings)

        self._update_markers_visible()
        self._update_line_visible()
        self._update_xerr_visible()
        self._update_yerr_visible()
        self._update_vectors_visible()

        self._update_size_mode()
        self._update_vector_mode()
        self._update_cmap_mode()

        self._update_checkboxes()

        self._update_warnings()

    def _update_warnings(self, *args):

        if self.layer_state.layer is None:
            n_points = 0
        else:
            n_points = np.product(self.layer_state.layer.shape)

        warning = " (may be slow given data size)"

        for combo, threshold in [(self.ui.combosel_size_mode, 10000),
                                 (self.ui.combosel_cmap_mode, 50000)]:

            if n_points > threshold and not self.layer_state.density_map:
                for item in range(combo.count()):
                    text = combo.itemText(item)
                    if text != 'Fixed':
                        combo.setItemText(item, text + warning)
                        combo.setItemData(item, QtGui.QBrush(Qt.red), Qt.TextColorRole)
            else:
                for item in range(combo.count()):
                    text = combo.itemText(item)
                    if text != 'Fixed':
                        if warning in text:
                            combo.setItemText(item, text.replace(warning, ''))
                            combo.setItemData(item, QtGui.QBrush(), Qt.TextColorRole)

        if n_points > 10000:
            self.ui.label_warning_errorbar.show()
        else:
            self.ui.label_warning_errorbar.hide()

        if n_points > 10000:
            self.ui.label_warning_vector.show()
        else:
            self.ui.label_warning_vector.hide()

    def _update_size_mode(self, size_mode=None):

        visible = not self.layer_state.density_map and not self.layer_state.size_mode == 'Fixed'
        self.ui.label_size_attribute.setVisible(visible)
        self.ui.combosel_size_att.setVisible(visible)
        self.ui.label_size_limits.setVisible(visible)
        self.ui.valuetext_size_vmin.setVisible(visible)
        self.ui.valuetext_size_vmax.setVisible(visible)
        self.ui.button_flip_size.setVisible(visible)

        visible = not self.layer_state.density_map and self.layer_state.size_mode == 'Fixed'
        self.ui.value_size.setVisible(visible)

        density = self.layer_state.density_map
        self.ui.value_dpi.setVisible(density)
        self.ui.label_dpi.setVisible(density)
        self.ui.label_stretch.setVisible(density)
        self.ui.combosel_stretch.setVisible(density)
        self.ui.value_density_contrast.setVisible(density)
        self.ui.label_contrast.setVisible(density)
        self.ui.combosel_size_mode.setVisible(not density)
        self.ui.value_size_scaling.setVisible(not density)
        self.ui.label_size_mode.setVisible(not density)
        self.ui.label_size_scaling.setVisible(not density)
        self.ui.label_fill.setVisible(not density)
        self.ui.bool_fill.setVisible(not density)

    def _update_markers_visible(self, *args):
        self.ui.combosel_size_mode.setEnabled(self.layer_state.markers_visible)
        self.ui.value_size.setEnabled(self.layer_state.markers_visible)
        self.ui.combosel_size_att.setEnabled(self.layer_state.markers_visible)
        self.ui.valuetext_size_vmin.setEnabled(self.layer_state.markers_visible)
        self.ui.valuetext_size_vmax.setEnabled(self.layer_state.markers_visible)
        self.ui.button_flip_size.setEnabled(self.layer_state.markers_visible)
        self.ui.value_size_scaling.setEnabled(self.layer_state.markers_visible)
        self.ui.value_dpi.setEnabled(self.layer_state.markers_visible)
        self.ui.combosel_stretch.setEnabled(self.layer_state.markers_visible)
        self.ui.label_size_scaling.setEnabled(self.layer_state.markers_visible)
        self.ui.combosel_points_mode.setEnabled(self.layer_state.markers_visible)
        self.ui.value_density_contrast.setEnabled(self.layer_state.markers_visible)

    def _update_checkboxes(self, *args):
        for checkbox in [self.ui.bool_line_visible, self.ui.bool_xerr_visible,
                         self.ui.bool_yerr_visible, self.ui.bool_vector_visible]:
            if self.layer_state.density_map:
                checkbox.setEnabled(False)
                checkbox.setToolTip('Not available with density map')
            else:
                checkbox.setEnabled(True)
                checkbox.setToolTip('')

    def _update_line_visible(self, *args):
        self.ui.value_linewidth.setEnabled(self.layer_state.line_visible)
        self.ui.combosel_linestyle.setEnabled(self.layer_state.line_visible)

    def _update_xerr_visible(self, *args):
        self.ui.combosel_xerr_att.setEnabled(self.layer_state.xerr_visible)

    def _update_yerr_visible(self, *args):
        self.ui.combosel_yerr_att.setEnabled(self.layer_state.yerr_visible)

    def _update_vectors_visible(self, *args):
        self.ui.combosel_vector_mode.setEnabled(self.layer_state.vector_visible)
        self.ui.combosel_vx_att.setEnabled(self.layer_state.vector_visible)
        self.ui.combosel_vy_att.setEnabled(self.layer_state.vector_visible)
        self.ui.value_vector_scaling.setEnabled(self.layer_state.vector_visible)
        self.ui.combosel_vector_origin.setEnabled(self.layer_state.vector_visible)
        self.ui.bool_vector_arrowhead.setEnabled(self.layer_state.vector_visible)

    def _update_vector_mode(self, vector_mode=None):
        if self.layer_state.vector_mode == 'Cartesian':
            self.ui.label_vector_x.setText('vx')
            self.ui.label_vector_y.setText('vy')
        elif self.layer_state.vector_mode == 'Polar':
            self.ui.label_vector_x.setText('angle (deg)')
            self.ui.label_vector_y.setText('length')

    def _update_cmap_mode(self, cmap_mode=None):

        if self.layer_state.cmap_mode == 'Fixed':
            self.ui.label_cmap_attribute.hide()
            self.ui.combosel_cmap_att.hide()
            self.ui.label_cmap_limits.hide()
            self.ui.valuetext_cmap_vmin.hide()
            self.ui.valuetext_cmap_vmax.hide()
            self.ui.button_flip_cmap.hide()
            self.ui.combodata_cmap.hide()
            self.ui.label_colormap.hide()
            self.ui.color_color.show()
        else:
            self.ui.label_cmap_attribute.show()
            self.ui.combosel_cmap_att.show()
            self.ui.label_cmap_limits.show()
            self.ui.valuetext_cmap_vmin.show()
            self.ui.valuetext_cmap_vmax.show()
            self.ui.button_flip_cmap.show()
            self.ui.combodata_cmap.show()
            self.ui.label_colormap.show()
            self.ui.color_color.hide()



from __future__ import absolute_import, division, print_function

from numpy.testing import assert_equal

from glue.core import Data, DataCollection, ComponentID
from glue.core.component_link import ComponentLink
from glue.core.parse import ParsedCommand, ParsedComponentLink

from ..component_manager import ComponentManagerWidget
from glue.dialogs.component_arithmetic.qt.tests.test_component_arithmetic import ChangeListener


class TestComponentManagerWidget:

    def setup_method(self):

        self.data1 = Data(x=[1, 2, 3], y=[3.5, 4.5, -1.0], z=['a', 'r', 'w'])
        self.data2 = Data(a=[3, 4, 1], b=[1.5, -2.0, 3.5], c=['y', 'e', 'r'])

        # Add a derived component so that we can test how we deal with existing ones
        components = dict((cid.label, cid) for cid in self.data2.components)
        pc = ParsedCommand('{a}', components)
        link = ParsedComponentLink(ComponentID('d'), pc)
        self.data2.add_component_link(link)

        self.data_collection = DataCollection([self.data1, self.data2])

        link = ComponentLink([self.data1.id['x']], self.data2.id['a'])
        self.data_collection.add_link(link)

        self.listener1 = ChangeListener(self.data1)
        self.listener2 = ChangeListener(self.data2)

    def test_nochanges(self):
        self.manager = ComponentManagerWidget(self.data_collection)
        self.manager.show()
        self.manager.button_ok.click()
        self.listener1.assert_exact_changes()
        self.listener2.assert_exact_changes()

    def test_remove(self):
        x_cid = self.data1.id['x']
        self.manager = ComponentManagerWidget(self.data_collection)
        self.manager.show()
        item = list(self.manager.list)[0]
        self.manager.list.select_item(item)
        self.manager.button_remove_main.click()
        self.manager.button_ok.click()
        self.listener1.assert_exact_changes(removed=[x_cid])
        self.listener2.assert_exact_changes()

    def test_rename_valid(self):
        x_cid = self.data1.id['x']
        self.manager = ComponentManagerWidget(self.data_collection)
        self.manager.show()
        item = list(self.manager.list)[0]
        item.setText(0, 'newname')
        self.manager.button_ok.click()
        assert self.manager.result() == 1
        self.listener1.assert_exact_changes(renamed=[x_cid])
        self.listener2.assert_exact_changes()
        assert x_cid.label == 'newname'
        assert_equal(self.data1['newname'], [1, 2, 3])

    def test_rename_invalid(self):
        x_cid = self.data1.id['x']
        self.manager = ComponentManagerWidget(self.data_collection)
        self.manager.show()
        item = list(self.manager.list)[0]
        item.setText(0, 'y')
        assert not self.manager.button_ok.isEnabled()
        assert self.manager.ui.label_status.text() == 'Error: some components have duplicate names'
        item = list(self.manager.list)[0]
        item.setText(0, 'a')
        assert self.manager.button_ok.isEnabled()
        assert self.manager.ui.label_status.text() == ''
        self.manager.button_ok.click()
        self.listener1.assert_exact_changes(renamed=[x_cid])
        self.listener2.assert_exact_changes()
        assert x_cid.label == 'a'
        assert_equal(self.data1['a'], [1, 2, 3])



from __future__ import absolute_import, division, print_function

from collections import OrderedDict
import numpy as np

from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QPainter, QTransform, QPen
from qtpy.QtWidgets import (QGraphicsView, QGraphicsScene, QApplication,
                            QGraphicsTextItem, QGraphicsEllipseItem,
                            QGraphicsLineItem)

from glue.utils.qt import mpl_to_qt_color, qt_to_mpl_color

COLOR_SELECTED = (0.2, 0.9, 0.2)
COLOR_CONNECTED = (0.6, 0.9, 0.9)
COLOR_DISCONNECTED = (0.9, 0.6, 0.6)


def get_pen(color, linewidth=1):
    color = mpl_to_qt_color(color)
    return QPen(color, linewidth, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin)


class Edge(QGraphicsLineItem):

    def __init__(self, node_source, node_dest, linewidth=3, zindex=5):
        self.linewidth = linewidth
        self.node_source = node_source
        self.node_dest = node_dest
        super(Edge, self).__init__(0, 0, 1, 1)
        self.setZValue(zindex)
        self.color = '0.5'

    def update_position(self):
        x0, y0 = self.node_source.node_position
        x1, y1 = self.node_dest.node_position
        self.setLine(x0, y0, x1, y1)

    @property
    def color(self):
        return qt_to_mpl_color(self.pen().color())

    @color.setter
    def color(self, value):
        self.setPen(get_pen(value, self.linewidth))

    def add_to_scene(self, scene):
        scene.addItem(self)

    def remove_from_scene(self, scene):
        scene.removeItem(self)

    def contains(self, point):
        return super(Edge, self).contains(self.mapFromScene(point))


class DataNode:

    def __init__(self, data, radius=15):

        self.data = data

        # Add circular node
        self.node = QGraphicsEllipseItem(0, 0, 1, 1)

        # Set radius
        self.radius = radius

        # Add text label
        self.label = QGraphicsTextItem(data.label)
        font = self.label.font()
        font.setPointSize(10)
        self.label.setFont(font)

        # Add line between label and node
        self.line1 = QGraphicsLineItem(0, 0, 1, 1)
        self.line2 = QGraphicsLineItem(0, 0, 1, 1)

        self.node.setZValue(20)
        self.label.setZValue(10)
        self.line1.setZValue(10)
        self.line2.setZValue(10)

        self.line1.setPen(get_pen('0.5'))
        self.line2.setPen(get_pen('0.5'))

        self.color = '0.8'

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value
        self.node.setRect(-value, -value, 2 * value, 2 * value)

    def contains(self, point):

        # Check label
        if self.label.contains(self.label.mapFromScene(point)):
            return True

        # Check node
        if self.node.contains(self.node.mapFromScene(point)):
            return True

        return False

    def update(self):
        self.node.update()

    def add_to_scene(self, scene):
        scene.addItem(self.node)
        scene.addItem(self.label)
        scene.addItem(self.line1)
        scene.addItem(self.line2)

    def remove_from_scene(self, scene):
        scene.removeItem(self.node)
        scene.removeItem(self.label)
        scene.removeItem(self.line1)
        scene.removeItem(self.line2)

    @property
    def node_position(self):
        pos = self.node.pos()
        return pos.x(), pos.y()

    @node_position.setter
    def node_position(self, value):
        self.node.setPos(value[0], value[1])
        self.update_lines()

    @property
    def label_position(self):
        pos = self.label.pos()
        return pos.x(), pos.y()

    @label_position.setter
    def label_position(self, value):
        self.label.setPos(value[0], value[1])
        self.update_lines()

    def update_lines(self):
        x0, y0 = self.label_position
        x2, y2 = self.node_position
        x1 = 0.5 * (x0 + x2)
        y1 = y0
        self.line1.setLine(x0, y0, x1, y1)
        self.line2.setLine(x1, y1, x2, y2)

    @property
    def color(self):
        return qt_to_mpl_color(self.node.brush().color())

    @color.setter
    def color(self, value):
        self.node.setBrush(mpl_to_qt_color(value))


def get_connections(dc_links):
    links = []
    for link in dc_links:
        data1 = link.data1
        data2 = link.data2
        if (data1, data2) not in links and (data2, data1) not in links:
            links.append((data1, data2))

    return links


def layout_simple_circle(nodes, edges, center=None, radius=None, reorder=True):

    # Place nodes around a circle

    if reorder:
        nodes[:] = order_nodes_by_connections(nodes, edges)

    for i, node in enumerate(nodes):
        angle = 2 * np.pi * i / len(nodes)
        nx = radius * np.cos(angle) + center[0]
        ny = radius * np.sin(angle) + center[1]
        node.node_position = nx, ny


def order_nodes_by_connections(nodes, edges):

    search_nodes = list(nodes)
    sorted_nodes = []

    while len(search_nodes) > 0:

        lengths = []
        connections = []

        for node in search_nodes:
            direct, indirect = find_connections(node, search_nodes, edges)
            connections.append((indirect, direct))
            lengths.append((len(indirect), len(direct)))

        m = max(lengths)

        for i in range(len(lengths)):
            if lengths[i] == m:
                for node in connections[i][0] + connections[i][1]:
                    if node not in sorted_nodes:
                        sorted_nodes.append(node)

        search_nodes = [node for node in nodes if node not in sorted_nodes]

    return sorted_nodes


class DataGraphWidget(QGraphicsView):

    selection_changed = Signal()

    def __init__(self, parent=None):

        super(DataGraphWidget, self).__init__(parent=parent)

        # Set up scene

        self.scene = QGraphicsScene(self)
        self.scene.setItemIndexMethod(QGraphicsScene.NoIndex)
        self.scene.setSceneRect(0, 0, 800, 300)

        self.setScene(self.scene)

        self.setWindowTitle("Glue data graph")

        self.setRenderHint(QPainter.Antialiasing)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QGraphicsView.AnchorViewCenter)

        self.selection_level = 0

    def resizeEvent(self, event):
        self.scene.setSceneRect(0, 0, self.width(), self.height())
        self.relayout(reorder=False)

    def relayout(self, reorder=True):

        # Update radius
        for node in self.nodes:
            node.radius = self.height() / 30.

        layout_simple_circle(self.nodes, self.edges,
                             center=(self.width() / 2, self.height() / 2),
                             radius=self.height() / 3, reorder=reorder)

        # Update edge positions
        for edge in self.background_edges + self.edges:
            edge.update_position()

        # Set up labels
        self.left_nodes = [node for node in self.nodes if node.node_position[0] < self.width() / 2]
        self.left_nodes = sorted(self.left_nodes, key=lambda x: x.node_position[1], reverse=True)

        self.right_nodes = [node for node in self.nodes if node not in self.left_nodes]
        self.right_nodes = sorted(self.right_nodes, key=lambda x: x.node_position[1], reverse=True)

        for i, node in enumerate(self.left_nodes):
            y = self.height() - (i + 1) / (len(self.left_nodes) + 1) * self.height()
            node.label_position = self.width() / 2 - self.height() / 2, y

        for i, node in enumerate(self.right_nodes):
            y = self.height() - (i + 1) / (len(self.right_nodes) + 1) * self.height()
            node.label_position = self.width() / 2 + self.height() / 2, y

    def set_data_collection(self, data_collection, old_links=None, new_links=None):

        # Get data and initialize nodes
        self.data_to_nodes = OrderedDict((data, DataNode(data)) for data in data_collection)
        self.nodes = list(self.data_to_nodes.values())

        # Get links and set up edges

        if old_links:
            self.background_edges = [Edge(self.data_to_nodes[data1], self.data_to_nodes[data2], linewidth=1, zindex=1)
                                     for data1, data2 in get_connections(data_collection.external_links)]
        else:
            self.background_edges = []

        if new_links:
            self.edges = [Edge(self.data_to_nodes[data1], self.data_to_nodes[data2])
                          for data1, data2 in get_connections(new_links)]
        else:
            self.edges = []

        # Figure out positions
        self.relayout()

        # Add nodes and edges to graph

        for node in self.nodes:
            node.add_to_scene(self.scene)

        for edge in self.background_edges + self.edges:
            edge.add_to_scene(self.scene)

        self.text_adjusted = False

        self.selected_edge = None
        self.selected_node1 = None
        self.selected_node2 = None

    def set_links(self, links):

        for edge in self.edges:
            edge.remove_from_scene(self.scene)

        self.edges = [Edge(self.data_to_nodes[data1], self.data_to_nodes[data2])
                      for data1, data2 in get_connections(links)]

        for edge in self.edges:
            edge.update_position()

        for edge in self.edges:
            edge.add_to_scene(self.scene)

        self._update_selected_edge()

        self._update_selected_colors()

    def paintEvent(self, event):

        super(DataGraphWidget, self).paintEvent(event)

        if not self.text_adjusted:

            for node in self.nodes:

                width = node.label.boundingRect().width()
                height = node.label.boundingRect().height()

                transform = QTransform()
                if node in self.left_nodes:
                    transform.translate(-width, -height / 2)
                else:
                    transform.translate(0, -height / 2)

                node.label.setTransform(transform)

            self.text_adjusted = True

    def manual_select(self, data1=None, data2=None):
        if data1 is None and data2 is not None:
            data1, data2 = data2, data1
        if data2 is not None:
            self.selection_level = 2
        elif data1 is not None:
            self.selection_level = 1
        else:
            self.selection_level = 0
        self.selected_node1 = self.data_to_nodes.get(data1, None)
        self.selected_node2 = self.data_to_nodes.get(data2, None)
        self._update_selected_edge()
        self._update_selected_colors()

    def find_object(self, event):
        for obj in list(self.nodes) + self.edges:
            if obj.contains(event.localPos()):
                return obj

    def mouseMoveEvent(self, event):

        # TODO: Don't update until the end
        # TODO: Only select object on top

        selected = self.find_object(event)

        if selected is None:

            if self.selection_level == 0:
                self.selected_node1 = None
                self.selected_node2 = None
                self._update_selected_edge()
            elif self.selection_level == 1:
                self.selected_node2 = None
                self._update_selected_edge()

        elif isinstance(selected, DataNode):

            if self.selection_level == 0:
                self.selected_node1 = selected
                self.selected_node2 = None
            elif self.selection_level == 1:
                if selected is not self.selected_node1:
                    self.selected_node2 = selected
                    self._update_selected_edge()

        elif isinstance(selected, Edge):

            if self.selection_level == 0:
                self.selected_edge = selected
                self.selected_node1 = selected.node_source
                self.selected_node2 = selected.node_dest

        self._update_selected_colors()

        self.selection_changed.emit()

    def mousePressEvent(self, event):

        # TODO: Don't update until the end
        # TODO: Only select object on top

        selected = self.find_object(event)

        if selected is None:

            self.selection_level = 0
            self.selected_node1 = None
            self.selected_node2 = None

            self._update_selected_edge()

        elif isinstance(selected, DataNode):

            if self.selection_level == 0:
                self.selected_node1 = selected
                self.selection_level += 1
            elif self.selection_level == 1:
                if selected is self.selected_node1:
                    self.selected_node1 = None
                    self.selection_level = 0
                else:
                    self.selected_node2 = selected
                    self.selection_level = 2
            elif self.selection_level == 2:
                if selected is self.selected_node2:
                    self.selected_node2 = None
                    self.selection_level = 1
                else:
                    self.selected_node1 = selected
                    self.selected_node2 = None
                    self.selection_level = 1

            self._update_selected_edge()

        elif isinstance(selected, Edge):

            if self.selected_edge is selected and self.selection_level == 2:
                self.selected_edge = None
                self.selected_node1 = None
                self.selected_node2 = None
                self.selection_level = 0
            else:
                self.selected_edge = selected
                self.selected_node1 = selected.node_source
                self.selected_node2 = selected.node_dest
                self.selection_level = 2

        self.mouseMoveEvent(event)

    def _update_selected_edge(self):
        for edge in self.edges:
            if (edge.node_source is self.selected_node1 and edge.node_dest is self.selected_node2 or
                    edge.node_source is self.selected_node2 and edge.node_dest is self.selected_node1):
                self.selected_edge = edge
                break
        else:
            self.selected_edge = None

    def _update_selected_colors(self):

        colors = {}

        if self.selected_node1 is not None and self.selection_level < 2:

            direct, indirect = find_connections(self.selected_node1, self.nodes, self.edges)

            for node in self.nodes:
                if node in direct or node in indirect:
                    colors[node] = COLOR_CONNECTED
                else:
                    colors[node] = COLOR_DISCONNECTED

            for edge in self.edges:
                if (edge.node_source is self.selected_node1 or
                        edge.node_dest is self.selected_node1):
                    colors[edge] = COLOR_CONNECTED

        if self.selected_edge is not None:
            colors[self.selected_edge] = COLOR_SELECTED

        if self.selected_node1 is not None:
            colors[self.selected_node1] = COLOR_SELECTED

        if self.selected_node2 is not None:
            colors[self.selected_node2] = COLOR_SELECTED

        self.set_colors(colors)

    def set_colors(self, colors):

        for obj in list(self.nodes) + self.edges:
            default_color = '0.8' if isinstance(obj, DataNode) else '0.5'
            obj.color = colors.get(obj, default_color)
            obj.update()


def find_connections(node, nodes, edges):

    direct = [node]
    indirect = []
    current = direct
    connected = [node]

    changed = True
    while changed:
        changed = False
        for edge in edges:
            source = edge.node_source
            dest = edge.node_dest
            if source in connected and dest not in connected:
                current.append(dest)
                changed = True
            if dest in connected and source not in connected:
                current.append(source)
                changed = True
        current = indirect
        connected.extend(current)
    return direct, indirect


if __name__ == '__main__':

    import sys

    app = QApplication(sys.argv)
    app.setAttribute(Qt.AA_UseHighDpiPixmaps)

    from glue.core.state import load
    dc = load('links.glu')

    widget = DataGraphWidget(dc)
    widget.show()

    sys.exit(app.exec_())



import sys

from . import CallbackProperty, HasCallbackProperties


class CallbackList(list):
    """
    A list that calls a callback function when it is modified.

    The first argument should be the callback function (which takes no
    arguments), and subsequent arguments are as for `list`.
    """

    def __init__(self, callback, *args, **kwargs):
        super(CallbackList, self).__init__(*args, **kwargs)
        self.callback = callback

    def __repr__(self):
        return "<CallbackList with {0} elements>".format(len(self))

    def append(self, value):
        super(CallbackList, self).append(value)
        if isinstance(value, HasCallbackProperties):
            value.add_global_callback(self.callback)
        self.callback()

    def extend(self, iterable):
        super(CallbackList, self).extend(iterable)
        for item in iterable:
            if isinstance(item, HasCallbackProperties):
                item.add_global_callback(self.callback)
        self.callback()

    def insert(self, index, value):
        super(CallbackList, self).insert(index, value)
        if isinstance(value, HasCallbackProperties):
            value.add_global_callback(self.callback)
        self.callback()

    def pop(self, index=-1):
        result = super(CallbackList, self).pop(index)
        if isinstance(result, HasCallbackProperties):
            result.remove_global_callback(self.callback)
        self.callback()
        return result

    def remove(self, value):
        if isinstance(value, HasCallbackProperties):
            value.remove_global_callback(self.callback)
        super(CallbackList, self).remove(value)
        self.callback()

    def reverse(self):
        super(CallbackList, self).reverse()
        self.callback()

    def sort(self, key=None, reverse=False):
        super(CallbackList, self).sort(key=key, reverse=reverse)
        self.callback()

    def __setitem__(self, slc, new_value):

        old_values = self[slc]
        if not isinstance(slc, slice):
            old_values = [old_values]

        for old_value in old_values:
            if isinstance(old_value, HasCallbackProperties):
                old_value.remove_global_callback(self.callback)

        if isinstance(slc, slice):
            new_values = new_value
        else:
            new_values = [new_value]

        for value in new_values:
            if isinstance(value, HasCallbackProperties):
                value.add_global_callback(self.callback)

        super(CallbackList, self).__setitem__(slc, new_value)
        self.callback()

    if sys.version_info[0] >= 3:

        def clear(self):
            for item in self:
                if isinstance(item, HasCallbackProperties):
                    item.remove_global_callback(self.callback)
            super(CallbackList, self).clear()
            self.callback()

    else:

        def __setslice__(self, start, end, new_values):

            slc = slice(start, end)

            old_values = self[slc]

            for old_value in old_values:
                if isinstance(old_value, HasCallbackProperties):
                    old_value.remove_global_callback(self.callback)

            for value in new_values:
                if isinstance(value, HasCallbackProperties):
                    value.add_global_callback(self.callback)

            super(CallbackList, self).__setslice__(start, end, new_values)
            self.callback()


class ListCallbackProperty(CallbackProperty):
    """
    A list property that calls callbacks when its contents are modified
    """

    def _default_getter(self, instance, owner=None):
        if instance not in self._values:
            self._default_setter(instance, [])
        return super(ListCallbackProperty, self)._default_getter(instance, owner)

    def _default_setter(self, instance, value):

        if not isinstance(value, list):
            raise TypeError('callback property should be a list')

        def callback(*args, **kwargs):
            self.notify(instance, wrapped_list, wrapped_list)

        wrapped_list = CallbackList(callback, value)
        super(ListCallbackProperty, self)._default_setter(instance, wrapped_list)
